from pathlib import Path

from propan.cli.startproject.async_app.core import create_app_file, create_handlers_file
from propan.cli.startproject.core import (
    create_apps_dir,
    create_config_dir,
    create_env,
    create_project_dir,
)
from propan.cli.startproject.utils import touch_dir, write_file


def create_redis(dir: Path) -> Path:
    project_dir = _create_project_dir(dir)
    app_dir = _create_app_dir(project_dir / "app")
    _create_config_dir(app_dir / "config")
    _create_apps_dir(app_dir / "apps")
    return project_dir


def _create_project_dir(dirname: Path) -> Path:
    project_dir = create_project_dir(dirname, "propan[async-redis]")

    write_file(
        project_dir / "docker-compose.yaml",
        'version: "3"',
        "",
        "services:",
        "  redis:",
        "    image: redis",
        "    healthcheck:",
        "      test: ['CMD', 'redis-cli', 'ping']",
        "      interval: 3s",
        "      timeout: 30s",
        "      retries: 3",
        "    ports:",
        "      - 6379:6379",
        "",
        "  app:",
        "    build: .",
        "    environment:",
        "      APP_BROKER__URL: redis://redis:6379/",
        "    volumes:",
        "      - ./app:/home/user/app:ro",
        "    depends_on:",
        "      redis:",
        "        condition: service_healthy",
    )

    return project_dir


def _create_app_dir(app: Path) -> Path:
    app_dir = touch_dir(app)
    create_app_file(app_dir, "RedisBroker")
    return app_dir


def _create_config_dir(config: Path) -> Path:
    config_dir = create_config_dir(config)
    create_env(config_dir, "redis://localhost:6379/")
    return config_dir


def _create_apps_dir(apps: Path) -> Path:
    apps_dir = create_apps_dir(apps)

    create_handlers_file(
        apps_dir / "handlers.py",
        "RedisRouter",
    )

    return apps_dir



from pathlib import Path


def touch_dir(dir: Path) -> Path:
    if not dir.exists():  # pragma: no branch
        dir.mkdir()
    return dir


def write_file(path: Path, *content: str) -> None:
    path.touch()
    if content:
        path.write_text("\n".join(content))



from typing import Any, Optional

from propan.brokers._model.schemas import PropanMessage
from propan.brokers.rabbit.schemas import RabbitExchange, RabbitQueue
from propan.types import AnyDict


class RabbitLoggingMixin:
    _max_queue_len: int
    _max_exchange_len: int

    def _get_log_context(
        self,
        message: Optional[PropanMessage[Any]],
        queue: RabbitQueue,
        exchange: Optional[RabbitExchange] = None,
    ) -> AnyDict:
        context = {
            "queue": queue.name,
            "exchange": exchange.name if exchange else "default",
            **super()._get_log_context(message),
        }
        return context

    @property
    def fmt(self) -> str:
        return super().fmt or (
            "%(asctime)s %(levelname)s - "
            f"%(exchange)-{self._max_exchange_len}s | "
            f"%(queue)-{self._max_queue_len}s | "
            f"%(message_id)-10s "
            "- %(message)s"
        )

    def _setup_log_context(
        self,
        queue: Optional[RabbitQueue] = None,
        exchange: Optional[RabbitExchange] = None,
    ) -> None:
        if exchange is not None:
            self._max_exchange_len = max(self._max_exchange_len, len(exchange.name))

        if queue is not None:  # pragma: no branch
            self._max_queue_len = max(self._max_queue_len, len(queue.name))



import asyncio
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from aiokafka import AIOKafkaConsumer
from fast_depends.core import CallModel

from propan.asyncapi.bindings import (
    AsyncAPIChannelBinding,
    AsyncAPIOperationBinding,
    kafka,
)
from propan.asyncapi.channels import AsyncAPIChannel
from propan.asyncapi.message import AsyncAPICorrelationId, AsyncAPIMessage
from propan.asyncapi.subscription import AsyncAPISubscription
from propan.brokers._model.schemas import BaseHandler
from propan.types import AnyDict, DecoratedCallable


@dataclass
class Handler(BaseHandler):
    topics: List[str]
    group_id: Optional[str] = None

    consumer: Optional[AIOKafkaConsumer] = None
    task: Optional["asyncio.Task[Any]"] = None
    consumer_kwargs: AnyDict = field(default_factory=dict)

    def __init__(
        self,
        callback: DecoratedCallable,
        dependant: CallModel,
        topics: List[str],
        group_id: Optional[str] = None,
        consumer: Optional[AIOKafkaConsumer] = None,
        task: Optional["asyncio.Task[Any]"] = None,
        consumer_kwargs: Optional[AnyDict] = None,
        _description: str = "",
    ):
        self.callback = callback
        self.dependant = dependant
        self._description = _description
        self.task = task
        self.topics = topics
        self.group_id = group_id
        self.consumer = consumer
        self.consumer_kwargs = consumer_kwargs or {}

    def get_schema(self) -> Dict[str, AsyncAPIChannel]:
        message_title, body, reply_to = self.get_message_object()

        if reply_to:
            kafka_kwargs = {"replyTo": reply_to}
        else:
            kafka_kwargs = {}

        if self.group_id is not None:
            kafka_kwargs["groupId"] = {"type": "string", "enum": [self.group_id]}

        if kafka_kwargs:
            operation_binding = AsyncAPIOperationBinding(
                kafka=kafka.AsyncAPIKafkaOperationBinding(**kafka_kwargs)  # type: ignore
            )
        else:
            operation_binding = None

        return {
            self.title: AsyncAPIChannel(
                subscribe=AsyncAPISubscription(
                    description=self.description,
                    bindings=operation_binding,
                    message=AsyncAPIMessage(
                        title=message_title,
                        payload=body,
                        correlationId=AsyncAPICorrelationId(
                            location="$message.header#/correlation_id"
                        ),
                    ),
                ),
                bindings=AsyncAPIChannelBinding(
                    kafka=kafka.AsyncAPIKafkaChannelBinding(
                        topic=self.topics,
                    )
                ),
            ),
        }



from pathlib import Path

from propan.cli.startproject.async_app.core import create_app_file, create_handlers_file
from propan.cli.startproject.core import (
    create_apps_dir,
    create_config_dir,
    create_env,
    create_project_dir,
)
from propan.cli.startproject.utils import touch_dir, write_file


def create_sqs(dir: Path) -> Path:
    project_dir = _create_project_dir(dir)
    app_dir = _create_app_dir(project_dir / "app")
    _create_config_dir(app_dir / "config")
    _create_apps_dir(app_dir / "apps")
    return project_dir


def _create_project_dir(dirname: Path) -> Path:
    project_dir = create_project_dir(dirname, "propan[async-redis]")

    write_file(
        project_dir / "docker-compose.yaml",
        'version: "3"',
        "",
        "services:",
        "  sqs:",
        "    image: softwaremill/elasticmq-native",
        "    ports:",
        "      - 9324:9324",
        "      - 9325:9325  # management",
        "",
        "  app:",
        "    build: .",
        "    environment:",
        "      APP_BROKER__URL: http://sqs:9324/",
        "    volumes:",
        "      - ./app:/home/user/app:ro",
        "    depends_on:",
        "      - sqs",
    )

    return project_dir


def _create_app_dir(app: Path) -> Path:
    app_dir = touch_dir(app)
    create_app_file(
        app_dir,
        "SQSBroker",
        imports=(
            "from botocore import UNSIGNED",
            "from aiobotocore.config import AioConfig",
        ),
        broker_init=(
            "    if settings.debug is True:",
            "        config = AioConfig(signature_version=UNSIGNED)",
            "    else:",
            "        config = None",
            "",
            "    await broker.connect(",
            "        settings.broker.url,",
            "        region_name=settings.broker.region,",
            "        config=config,",
            "    )",
        ),
    )
    return app_dir


def _create_config_dir(config: Path) -> Path:
    config_dir = create_config_dir(
        config,
        "class BrokerSettings(BaseModel):",
        "    url: str = Field(...)",
        '    region: str = "us-west-2"',
    )
    create_env(config_dir, "http://localhost:9324/")
    return config_dir


def _create_apps_dir(apps: Path) -> Path:
    apps_dir = create_apps_dir(apps)

    create_handlers_file(
        apps_dir / "handlers.py",
        "SQSRouter",
    )

    return apps_dir



import logging
from enum import Enum
from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Type, Union

from aiobotocore.config import AioConfig
from fastapi import params
from fastapi.datastructures import Default
from fastapi.routing import APIRoute
from fastapi.utils import generate_unique_id
from starlette import routing
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp

from propan import SQSBroker
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers.sqs.schema import SQSQueue
from propan.fastapi.core.router import PropanRouter
from propan.log import access_logger
from propan.types import AnyDict

class SQSRouter(PropanRouter[SQSBroker]):
    def __init__(
        self,
        url: str = "http://localhost:9324/",
        *,
        region_name: Optional[str] = None,
        api_version: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[bool] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        config: Optional[AioConfig] = None,
        # FastAPI kwargs
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[routing.BaseRoute]] = None,
        routes: Optional[List[routing.BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(
            generate_unique_id
        ),
        # Broker kwargs
        schema_url: str = "/asyncapi",
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        protocol: str = "sqs",
    ) -> None:
        pass
    def add_api_mq_route(  # type: ignore[override]
        self,
        queue: Union[str, SQSQueue],
        *,
        wait_interval: int = 1,
        max_messages_number: int = 10,  # 1...10
        attributes: Sequence[str] = (),
        message_attributes: Sequence[str] = (),
        request_attempt_id: Optional[str] = None,
        visibility_timeout: int = 0,
        retry: Union[bool, int] = False,
        endpoint: HandlerCallable[T_HandlerReturn],
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]]:
        pass
    def event(  # type: ignore[override]
        self,
        queue: Union[str, SQSQueue],
        *,
        wait_interval: int = 1,
        max_messages_number: int = 10,  # 1...10
        attributes: Sequence[str] = (),
        message_attributes: Sequence[str] = (),
        request_attempt_id: Optional[str] = None,
        visibility_timeout: int = 0,
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]:
        pass



import logging
import ssl
from enum import Enum
from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Type, Union

from fastapi import params
from fastapi.datastructures import Default
from fastapi.routing import APIRoute
from fastapi.utils import generate_unique_id
from nats.aio.client import (
    DEFAULT_CONNECT_TIMEOUT,
    DEFAULT_DRAIN_TIMEOUT,
    DEFAULT_INBOX_PREFIX,
    DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
    DEFAULT_MAX_OUTSTANDING_PINGS,
    DEFAULT_MAX_RECONNECT_ATTEMPTS,
    DEFAULT_PENDING_SIZE,
    DEFAULT_PING_INTERVAL,
    DEFAULT_RECONNECT_TIME_WAIT,
    Callback,
    Credentials,
    ErrorCallback,
    JWTCallback,
    SignatureCallback,
)
from nats.aio.msg import Msg
from nats.js.client import (
    DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
    DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
)
from starlette import routing
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp

from propan import NatsBroker, NatsJSBroker
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers.nats import consts as api
from propan.fastapi.core.router import PropanRouter
from propan.log import access_logger

class NatsRouter(PropanRouter[NatsBroker]):
    def __init__(
        self,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
        # FastAPI kwargs
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[routing.BaseRoute]] = None,
        routes: Optional[List[routing.BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(
            generate_unique_id
        ),
        # Broker kwargs
        schema_url: str = "/asyncapi",
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        protocol: str = "nats",
    ) -> None:
        pass
    def add_api_mq_route(  # type: ignore[override]
        self,
        subject: str,
        *,
        queue: str = "",
        endpoint: HandlerCallable[T_HandlerReturn],
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[[Msg, bool], Awaitable[T_HandlerReturn]]:
        pass
    def event(  # type: ignore[override]
        self,
        subject: str,
        *,
        queue: str = "",
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Msg, bool], Awaitable[T_HandlerReturn]],
    ]:
        pass

class NatsJSRouter(PropanRouter[NatsJSBroker]):
    def __init__(
        self,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        *,
        # JS
        stream: Optional[str] = "propan-stream",
        description: Optional[str] = None,
        retention: Optional[api.RetentionPolicy] = None,
        max_consumers: Optional[int] = None,
        max_msgs: Optional[int] = None,
        max_bytes: Optional[int] = None,
        discard: Optional[api.DiscardPolicy] = api.DiscardPolicy.OLD,
        max_age: Optional[float] = None,  # in seconds
        max_msgs_per_subject: int = -1,
        max_msg_size: Optional[int] = -1,
        storage: Optional[api.StorageType] = None,
        num_replicas: Optional[int] = None,
        no_ack: bool = False,
        template_owner: Optional[str] = None,
        duplicate_window: int = 0,
        placement: Optional[api.Placement] = None,
        mirror: Optional[api.StreamSource] = None,
        sources: Optional[List[api.StreamSource]] = None,
        sealed: bool = False,
        deny_delete: bool = False,
        deny_purge: bool = False,
        allow_rollup_hdrs: bool = False,
        republish: Optional[api.RePublish] = None,
        allow_direct: Optional[bool] = None,
        mirror_direct: Optional[bool] = None,
        # connection
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
        # FastAPI kwargs
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[routing.BaseRoute]] = None,
        routes: Optional[List[routing.BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(
            generate_unique_id
        ),
        # Broker kwargs
        schema_url: str = "/asyncapi",
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        protocol: str = "nats",
    ) -> None:
        pass
    def add_api_mq_route(  # type: ignore[override]
        self,
        subject: str,
        *,
        durable: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        ordered_consumer: bool = False,
        idle_heartbeat: Optional[float] = None,
        flow_control: bool = False,
        pending_msgs_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        deliver_policy: Optional[api.DeliverPolicy] = None,
        headers_only: Optional[bool] = None,
        # fastapi kwargs
        endpoint: HandlerCallable[T_HandlerReturn],
        # broker kwargs
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[[Msg, bool], Awaitable[T_HandlerReturn]]:
        pass
    def event(  # type: ignore[override]
        self,
        subject: str,
        *,
        queue: str = "",
        durable: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        ordered_consumer: bool = False,
        idle_heartbeat: Optional[float] = None,
        flow_control: bool = False,
        pending_msgs_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        deliver_policy: Optional[api.DeliverPolicy] = None,
        headers_only: Optional[bool] = None,
        # broker kwargs
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Msg, bool], Awaitable[T_HandlerReturn]],
    ]:
        pass



from contextlib import asynccontextmanager
from functools import wraps
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from uuid import uuid4

import aio_pika
import aiormq
import anyio
from aio_pika.abc import DeliveryMode
from anyio.streams.memory import MemoryObjectReceiveStream
from fast_depends.dependencies import Depends
from typing_extensions import TypeAlias
from yarl import URL

from propan._compat import model_to_dict
from propan.brokers._model.broker_usecase import (
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.exceptions import WRONG_PUBLISH_ARGS
from propan.brokers.push_back_watcher import BaseWatcher, WatcherContext
from propan.brokers.rabbit.logging import RabbitLoggingMixin
from propan.brokers.rabbit.schemas import Handler, RabbitExchange, RabbitQueue
from propan.brokers.rabbit.utils import validate_exchange, validate_queue
from propan.types import AnyDict, SendableMessage
from propan.utils import context

TimeoutType = Optional[Union[int, float]]
PikaSendableMessage: TypeAlias = Union[aio_pika.message.Message, SendableMessage]
RabbitMessage: TypeAlias = PropanMessage[aio_pika.message.IncomingMessage]
T = TypeVar("T")

RABBIT_REPLY = "amq.rabbitmq.reply-to"


class RabbitBroker(
    RabbitLoggingMixin,
    BrokerAsyncUsecase[aio_pika.message.IncomingMessage, aio_pika.RobustConnection],
):
    handlers: List[Handler]
    _connection: Optional[aio_pika.RobustConnection]
    _channel: Optional[aio_pika.RobustChannel]
    _queues: Dict[RabbitQueue, aio_pika.RobustQueue]
    _exchanges: Dict[RabbitExchange, aio_pika.RobustExchange]
    _rpc_lock: anyio.Lock

    def __init__(
        self,
        url: Union[str, URL, None] = None,
        *,
        log_fmt: Optional[str] = None,
        consumers: Optional[int] = None,
        protocol: str = "amqp",
        protocol_version: str = "0.9.1",
        **kwargs: AnyDict,
    ) -> None:
        super().__init__(
            url,
            log_fmt=log_fmt,
            url_=url or "amqp://guest:guest@localhost:5672/",
            protocol=protocol,
            protocol_version=protocol_version,
            **kwargs,
        )
        self._max_consumers = consumers

        self._channel = None
        self._rpc_lock = anyio.Lock()

        self._max_queue_len = 4
        self._max_exchange_len = 4
        self._queues = {}
        self._exchanges = {}

    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        await super().close(exc_type, exc_val, exec_tb)
        if self._channel is not None:
            await self._channel.close()
            self._channel = None

        if self._connection is not None:
            await self._connection.close()
            self._connection = None

        self._queues = {}
        self._exchanges = {}

    async def _connect(
        self,
        **kwargs: Any,
    ) -> aio_pika.RobustConnection:
        connection = await aio_pika.connect_robust(**kwargs)

        if self._channel is None:  # pragma: no branch
            max_consumers = self._max_consumers
            self._channel = await connection.channel()

            if max_consumers:
                c = self._get_log_context(None, RabbitQueue(""), RabbitExchange(""))
                self._log(f"Set max consumers to {max_consumers}", extra=c)
                await self._channel.set_qos(prefetch_count=int(max_consumers))

        return connection

    def handle(
        self,
        queue: Union[str, RabbitQueue],
        exchange: Union[str, RabbitExchange, None] = None,
        *,
        consume_arguments: Optional[AnyDict] = None,
        dependencies: Sequence[Depends] = (),
        description: str = "",
        **original_kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[aio_pika.message.IncomingMessage, bool], Awaitable[T_HandlerReturn]],
    ]:
        super().handle()

        queue, exchange = validate_queue(queue), validate_exchange(exchange)

        self._setup_log_context(queue, exchange)

        def wrapper(
            func: HandlerCallable[T_HandlerReturn],
        ) -> Callable[
            [aio_pika.message.IncomingMessage, bool], Awaitable[T_HandlerReturn]
        ]:
            func, dependant = self._wrap_handler(
                func,
                queue=queue,
                exchange=exchange,
                extra_dependencies=dependencies,
                **original_kwargs,
            )
            handler = Handler(
                callback=func,
                queue=queue,
                exchange=exchange,
                _description=description,
                consume_arguments=consume_arguments,
                dependant=dependant,
            )
            self.handlers.append(handler)

            return func

        return wrapper

    async def start(self) -> None:
        context.set_local(
            "log_context",
            self._get_log_context(None, RabbitQueue(""), RabbitExchange("")),
        )

        await super().start()

        for handler in self.handlers:
            queue = await self._init_handler(handler)

            func = handler.callback

            c = self._get_log_context(None, handler.queue, handler.exchange)
            self._log(f"`{func.__name__}` waiting for messages", extra=c)

            await queue.consume(func, arguments=handler.consume_arguments)

    async def publish(
        self,
        message: PikaSendableMessage = "",
        queue: Union[RabbitQueue, str] = "",
        exchange: Union[RabbitExchange, str, None] = None,
        *,
        routing_key: str = "",
        mandatory: bool = True,
        immediate: bool = False,
        timeout: TimeoutType = None,
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
        persist: bool = False,
        reply_to: Optional[str] = None,
        **message_kwargs,
    ) -> Union[aiormq.abc.ConfirmationFrameType, Dict, str, bytes, None]:
        if self._channel is None:
            raise ValueError("RabbitBroker channel not started yet")

        queue, exchange = validate_queue(queue), validate_exchange(exchange)

        if callback is True:
            if reply_to is not None:
                raise WRONG_PUBLISH_ARGS
            else:
                context = RPCCallback(self._rpc_lock, self)
        else:
            context = fake_context(reply_to)

        if exchange is None:
            exchange_obj = self._channel.default_exchange
        else:
            exchange_obj = await self.declare_exchange(exchange)

        async with context as (reply_to, response_queue):
            message = self._validate_message(
                message=message,
                persist=persist,
                reply_to=reply_to,
                **message_kwargs,
            )

            r = await exchange_obj.publish(
                message=message,
                routing_key=routing_key or queue.routing or "",
                mandatory=mandatory,
                immediate=immediate,
                timeout=timeout,
            )

            if response_queue is None:
                return r

            else:
                if raise_timeout:
                    scope = anyio.fail_after
                else:
                    scope = anyio.move_on_after

                msg: Any = None
                with scope(callback_timeout):
                    msg = await response_queue.receive()

                if msg:
                    return await self._decode_message(msg)

    async def _init_handler(
        self,
        handler: Handler,
    ) -> aio_pika.abc.AbstractRobustQueue:
        queue = await self.declare_queue(handler.queue)
        if handler.exchange is not None and handler.exchange.name != "default":
            exchange = await self.declare_exchange(handler.exchange)
            await queue.bind(
                exchange,
                routing_key=handler.queue.routing,
                arguments=handler.queue.bind_arguments,
            )
        return queue

    async def declare_queue(self, queue: RabbitQueue) -> aio_pika.RobustQueue:
        q = self._queues.get(queue)
        if q is None:
            q = await self._channel.declare_queue(**model_to_dict(queue))
            self._queues[queue] = q
        return q

    async def declare_exchange(
        self, exchange: RabbitExchange
    ) -> aio_pika.RobustExchange:
        exch = self._exchanges.get(exchange)

        if exch is None:
            exch = await self._channel.declare_exchange(**model_to_dict(exchange))
            self._exchanges[exchange] = exch

        if exchange.bind_to is not None:
            parent = await self.declare_exchange(exchange.bind_to)
            await exch.bind(
                exchange=parent,
                routing_key=exchange.routing_key,
                arguments=exchange.bind_arguments,
            )

        return exch

    @staticmethod
    async def _parse_message(
        message: aio_pika.message.IncomingMessage,
    ) -> RabbitMessage:
        return PropanMessage(
            body=message.body,
            headers=message.headers,
            reply_to=message.reply_to or "",
            message_id=message.message_id,
            content_type=message.content_type or "",
            raw_message=message,
        )

    def _process_message(
        self,
        func: Callable[[RabbitMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[RabbitMessage], Awaitable[T]]:
        @wraps(func)
        async def wrapper(message: RabbitMessage) -> T:
            pika_message = message.raw_message
            if watcher is None:
                context = pika_message.process()
            else:
                context = WatcherContext(
                    watcher,
                    message,
                    on_success=ack,
                    on_error=nack,
                    on_max=reject,
                )

            async with context:
                r = await func(message)
                if message.reply_to:
                    await self.publish(
                        message=r,
                        routing_key=message.reply_to,
                        correlation_id=pika_message.correlation_id,
                    )

                return r

        return wrapper

    @classmethod
    def _validate_message(
        cls: Type["RabbitBroker"],
        message: PikaSendableMessage,
        persist: bool = False,
        callback_queue: Optional[aio_pika.abc.AbstractRobustQueue] = None,
        reply_to: Optional[str] = None,
        **message_kwargs: Dict[str, Any],
    ) -> aio_pika.Message:
        if not isinstance(message, aio_pika.message.Message):
            message, content_type = super()._encode_message(message)

            delivery_mode = (
                DeliveryMode.PERSISTENT if persist else DeliveryMode.NOT_PERSISTENT
            )

            message = aio_pika.Message(
                message,
                **{
                    "delivery_mode": delivery_mode,
                    "content_type": content_type,
                    "reply_to": callback_queue or reply_to,
                    "correlation_id": str(uuid4()),
                    **message_kwargs,
                },
            )

        return message

    @property
    def channel(self) -> aio_pika.RobustChannel:
        return self._channel


class RPCCallback:
    def __init__(self, lock: anyio.Lock, broker: RabbitBroker):
        self.lock = lock
        self.broker = broker

    async def __aenter__(self) -> Tuple[str, MemoryObjectReceiveStream]:
        (
            send_response_stream,
            receive_response_stream,
        ) = anyio.create_memory_object_stream(max_buffer_size=1)
        await self.lock.acquire()

        self.queue = callback_queue = await self.broker._channel.get_queue(RABBIT_REPLY)

        async def handle_response(msg: RabbitMessage) -> None:
            propan_message = await self.broker._parse_message(msg)
            await send_response_stream.send(propan_message)

        self.consumer_tag = await callback_queue.consume(handle_response, no_ack=True)

        return callback_queue.name, receive_response_stream

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ):
        self.lock.release()
        await self.queue.cancel(self.consumer_tag)


@asynccontextmanager
async def fake_context(reply_to: str) -> Tuple[str, None]:
    yield reply_to, None


async def ack(message: RabbitMessage) -> None:
    pika_message = message.raw_message
    if (
        pika_message._IncomingMessage__processed
        or pika_message._IncomingMessage__no_ack
    ):
        return
    await pika_message.ack()


async def nack(message: RabbitMessage) -> None:
    pika_message = message.raw_message
    if (
        pika_message._IncomingMessage__processed
        or pika_message._IncomingMessage__no_ack
    ):
        return
    await pika_message.nack()


async def reject(message: RabbitMessage) -> None:
    pika_message = message.raw_message
    if (
        pika_message._IncomingMessage__processed
        or pika_message._IncomingMessage__no_ack
    ):
        return
    await pika_message.reject()



from typing import Optional

from pydantic import BaseModel, Field
from typing_extensions import Literal

from propan.types import AnyDict


class AsyncAPIRedisChannelBinding(BaseModel):
    channel: str
    method: Literal["ssubscribe", "psubscribe", "subscribe"] = "subscribe"
    version: str = Field(
        default="custom",
        alias="bindingVersion",
    )


class AsyncAPIRedisOperationBinding(BaseModel):
    reply_to: Optional[AnyDict] = Field(default=None, alias="replyTo")
    version: str = Field(
        default="custom",
        alias="bindingVersion",
    )



from typing import Awaitable, Callable

from aiokafka.structs import ConsumerRecord

from propan.brokers._model.broker_usecase import HandlerCallable, T_HandlerReturn
from propan.brokers._model.routing import BrokerRouter
from propan.types import AnyDict


class KafkaRouter(BrokerRouter[ConsumerRecord]):
    def handle(  # type: ignore[override]
        self,
        *topics: str,
        **kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[ConsumerRecord, bool], Awaitable[T_HandlerReturn]],
    ]:
        return super().handle(*[self.prefix + t for t in topics], **kwargs)



from typing import List, Optional

from pydantic import BaseModel

from propan.asyncapi.bindings import AsyncAPIChannelBinding, AsyncAPIOperationBinding
from propan.asyncapi.subscription import AsyncAPISubscription


class AsyncAPIPublish(BaseModel):
    bindings: Optional[AsyncAPIOperationBinding] = None


class AsyncAPIChannelParameters(BaseModel):
    # TODO
    ...


class AsyncAPIChannel(BaseModel):
    description: Optional[str] = None
    servers: Optional[List[str]] = None
    bindings: Optional[AsyncAPIChannelBinding] = None
    subscribe: Optional[AsyncAPISubscription] = None
    publish: Optional[AsyncAPIPublish] = None
    parameters: Optional[AsyncAPIChannelParameters] = None



from io import StringIO
from pathlib import Path
from typing import Any, Dict, cast

import typer

from propan._compat import model_to_jsonable
from propan.asyncapi import (
    AsyncAPIChannel,
    AsyncAPIComponents,
    AsyncAPIInfo,
    AsyncAPIMessage,
    AsyncAPISchema,
    AsyncAPIServer,
)
from propan.brokers._model import BrokerAsyncUsecase
from propan.cli.app import PropanApp
from propan.types import AnyDict


def generate_doc_file(app: PropanApp, filename: Path) -> None:
    schema = get_app_schema(app)
    json_schema = schema_to_json(schema)
    yaml_schema = json_schema_to_yaml(json_schema)
    filename.write_text(yaml_schema)
    typer.echo(f"Your project AsyncAPI scheme was placed to `{filename}`")


def gen_app_schema_yaml(app: PropanApp) -> str:
    json_schema = gen_app_schema_json(app)
    return json_schema_to_yaml(json_schema)


def gen_app_schema_json(app: PropanApp) -> AnyDict:
    schema = get_app_schema(app)
    return schema_to_json(schema)


def json_schema_to_yaml(schema: AnyDict) -> str:
    try:
        import yaml
    except ImportError as e:  # pragma: no cover
        typer.echo(
            "To generate documentation, please install the dependencies\n"
            'pip install "propan[doc]"'
        )
        raise typer.Exit(1) from e

    io = StringIO(initial_value="", newline="\n")
    yaml.dump(schema, io, sort_keys=False)
    return io.getvalue()


def schema_to_json(schema: AsyncAPISchema) -> AnyDict:
    return cast(
        AnyDict,
        model_to_jsonable(
            schema,
            by_alias=True,
            exclude_none=True,
        ),
    )


def get_app_schema(app: PropanApp) -> AsyncAPISchema:
    if not isinstance(app.broker, BrokerAsyncUsecase):
        raise typer.BadParameter("Your PropanApp broker is invalid")

    servers = _get_broker_servers(app.broker)

    messages: Dict[str, AsyncAPIMessage] = {}
    payloads: Dict[str, AnyDict] = {}

    channels = _get_broker_channels(app.broker)
    for channel_name, ch in channels.items():
        ch.servers = list(servers.keys())

        if ch.subscribe is not None:  # pragma: no branch
            m = ch.subscribe.message
            m_title = m.title or f"{channel_name}Message"

            p = m.payload
            p_title = p.get("title", f"{channel_name}Payload")
            payloads[p_title] = p

            m.payload = {"$ref": f"#/components/schemas/{p_title}"}

            messages[m_title] = m
            ch.subscribe.message = {
                "$ref": f"#/components/messages/{m_title}"
            }  # type: ignore

    schema = AsyncAPISchema(
        info=_get_app_info(app),
        servers=servers,
        channels=channels,
        components=AsyncAPIComponents(
            messages=messages,
            schemas=payloads,
        ),
    )
    return schema


def _get_app_info(app: PropanApp) -> AsyncAPIInfo:
    return AsyncAPIInfo(
        title=app.title,
        version=app.version,
        description=app.description,
        license=getattr(app, "license", None),
        contact=getattr(app, "contact", None),
    )


def _get_broker_servers(
    broker: BrokerAsyncUsecase[Any, Any]
) -> Dict[str, AsyncAPIServer]:
    if isinstance(broker.url, str):
        url = broker.url
    else:
        url = broker.url[0]

    return {
        "dev": AsyncAPIServer(
            url=url,
            protocol=broker.protocol,
            protocolVersion=broker.protocol_version,
        )
    }


def _get_broker_channels(
    broker: BrokerAsyncUsecase[Any, Any]
) -> Dict[str, AsyncAPIChannel]:
    channels = {}
    for handler in broker.handlers:
        channels.update(handler.get_schema())

    return channels



from typing import List, Optional

from pydantic import BaseModel, Field

from propan.types import AnyDict


class AsyncAPIKafkaChannelBinding(BaseModel):
    topic: List[str]
    partitions: Optional[int] = None
    replicas: Optional[int] = None
    # TODO:
    # topicConfiguration
    version: str = Field(
        default="0.4.0",
        alias="bindingVersion",
    )


class AsyncAPIKafkaOperationBinding(BaseModel):
    group_id: Optional[AnyDict] = Field(
        default=None,
        alias="groupId",
    )
    client_id: Optional[AnyDict] = Field(
        default=None,
        alias="clientId",
    )

    reply_to: Optional[AnyDict] = Field(default=None, alias="replyTo")

    version: str = Field(
        default="0.4.0",
        alias="bindingVersion",
    )



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Generic, Optional, Tuple, Type, TypeVar, Union
from uuid import uuid4

from fast_depends.core import CallModel
from pydantic import BaseModel, Field, Json, create_model
from pydantic.dataclasses import dataclass as pydantic_dataclass

from propan._compat import get_model_fileds, model_schema, update_model_example
from propan.asyncapi.channels import AsyncAPIChannel
from propan.types import AnyDict, DecodedMessage, DecoratedCallable


@dataclass
class BaseHandler:
    callback: DecoratedCallable
    dependant: CallModel
    _description: str

    def __init__(
        self,
        callback: DecoratedCallable,
        dependant: CallModel,
        _description: str = "",
    ):
        self.callback = callback
        self.dependant = dependant
        self._description = _description

    @abstractmethod
    def get_schema(self) -> Dict[str, AsyncAPIChannel]:
        raise NotImplementedError()

    @property
    def title(self) -> str:
        return self.callback.__name__.replace("_", " ").title().replace(" ", "")

    @property
    def description(self) -> Optional[str]:
        return self._description or self.callback.__doc__

    def get_message_object(self) -> Tuple[str, AnyDict, Optional[AnyDict]]:
        import jsonref  # hide it there to remove docs dependencies from main package

        dependant = self.dependant

        if getattr(dependant, "response_model", None) is not None:
            response_model: Type[BaseModel] = dependant.response_model
            return_field = list(get_model_fileds(response_model).values())[0]

            if (
                return_field.annotation != Any  # NOTE: 3.7-3.10 compatibility
                and issubclass(return_field.annotation, BaseModel)
            ):
                response_model = update_model_example(return_field.annotation)

                return_info = jsonref.replace_refs(
                    model_schema(response_model), jsonschema=True, proxies=False
                )
                return_info["examples"] = [return_info.pop("example", None)]

            else:
                response_model = update_model_example(response_model)
                response_field_name = "response"

                raw = jsonref.replace_refs(
                    model_schema(response_model),
                    jsonschema=True,
                    proxies=False,
                )

                return_info = raw.get("properties", {}).get(response_field_name)
                return_info["examples"] = [
                    raw.pop("example", {}).get(response_field_name)
                ]
                return_info["title"] = f"{self.title}Reply"
        else:
            return_info = None

        payload_title = f"{self.title}Payload"
        params = dependant.flat_params
        params_number = len(params)

        use_original_model = False
        if params_number == 0:
            model = None

        elif params_number == 1:
            name, param = list(params.items())[0]
            info = getattr(param, "field_info", param)

            if param.annotation != Any and issubclass(  # NOTE: 3.7-3.10 compatibility
                param.annotation, BaseModel
            ):
                model = param.annotation
                use_original_model = True

            else:
                model = create_model(  # type: ignore
                    info.title or payload_title,
                    **{name: (param.annotation, info)},
                )

        else:
            model = create_model(  # type: ignore
                payload_title,
                **{
                    i: (j.annotation, getattr(j, "field_info", j))
                    for i, j in params.items()
                },
            )

        body: AnyDict
        if model is None:
            body = {"title": payload_title, "type": "null"}
        else:
            model = update_model_example(model)
            body = jsonref.replace_refs(
                model_schema(model), jsonschema=True, proxies=False
            )

        body.pop("definitions", None)
        if return_info is not None:
            return_info.pop("definitions", None)

        if params_number == 1 and not use_original_model:
            param_body: AnyDict = body.get("properties", {})
            key = list(param_body.keys())[0]
            param_body = param_body[key]
            param_body.update(
                {
                    "example": body.get("example", {}).get(key),
                    "title": body.get("title", param_body.get("title")),
                }
            )
            param_body["example"] = body.get("example", {}).get(key)
            body = param_body

        return f"{self.title}Message", body, return_info


class NameRequired(BaseModel):
    name: Optional[str] = Field(...)

    def __eq__(self, __value: Optional["NameRequired"]) -> bool:
        return __value and self.name == __value.name

    def __init__(self, name: str, **kwargs: Any):
        super().__init__(name=name, **kwargs)


class Queue(NameRequired):
    name: Optional[str] = Field(...)

    def __init__(self, name: str, **kwargs: Any):
        super().__init__(name=name, **kwargs)


class RawDecoced(BaseModel):
    message: Union[Json[Any], str]


Msg = TypeVar("Msg")


@pydantic_dataclass
class PropanMessage(Generic[Msg]):
    raw_message: Msg
    body: Union[bytes, Any]
    content_type: Optional[str] = None
    reply_to: str = ""
    headers: AnyDict = Field(default_factory=dict)
    message_id: str = Field(default_factory=lambda: str(uuid4()))
    decoded_body: Optional[DecodedMessage] = None



from contextlib import contextmanager
from contextvars import ContextVar, Token
from typing import Any, Dict, Iterator, TypeVar

from propan.utils.classes import Singleton

T = TypeVar("T")


class ContextRepo(Singleton):
    _global_context: Dict[str, Any]
    _scope_context: Dict[str, ContextVar[Any]]

    def __init__(self) -> None:
        self._global_context = {}
        self._scope_context = {}

    def set_global(self, key: str, v: Any) -> None:
        self._global_context[key] = v

    def reset_global(self, key: str) -> None:
        self._global_context.pop(key, None)

    def set_local(self, key: str, value: T) -> "Token[T]":
        context_var = self._scope_context.get(key)
        if context_var is None:
            context_var = ContextVar(key, default=None)
            self._scope_context[key] = context_var
        return context_var.set(value)

    def reset_local(self, key: str, tag: "Token[Any]") -> None:
        self._scope_context[key].reset(tag)

    def get_local(self, key: str) -> Any:
        context_var = self._scope_context.get(key)
        if context_var is not None:  # pragma: no branch
            return context_var.get()

    def clear(self) -> None:
        self._global_context = {}
        self._scope_context = {}

    def get(self, key: str) -> Any:
        return self.context.get(key)

    def __getattr__(self, __name: str) -> Any:
        return self.get(__name)

    @property
    def context(self) -> Dict[str, Any]:
        return {
            "context": self,
            **{i: j.get() for i, j in self._scope_context.items()},
            **self._global_context,
        }

    @contextmanager
    def scope(self, key: str, value: Any) -> Iterator[None]:
        token = self.set_local(key, value)
        yield
        self.reset_local(key, token)


context: ContextRepo = ContextRepo()



from typing import Any, Awaitable, Callable, Union

from aio_pika.message import IncomingMessage

from propan._compat import model_copy
from propan.brokers._model.broker_usecase import HandlerCallable, T_HandlerReturn
from propan.brokers._model.routing import BrokerRouter
from propan.brokers.rabbit.schemas import RabbitQueue
from propan.brokers.rabbit.utils import validate_queue
from propan.types import AnyDict


class RabbitRouter(BrokerRouter[IncomingMessage]):
    def handle(
        self,
        queue: Union[str, RabbitQueue],
        *args: Any,
        **kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[IncomingMessage, bool], Awaitable[T_HandlerReturn]],
    ]:
        q = validate_queue(queue)
        return super().handle(model_copy(q, update={
            "name": self.prefix + q.name
        }), *args, **kwargs)



from dataclasses import dataclass
from typing import Dict, Optional

from fast_depends.core import CallModel
from nats.aio.subscription import Subscription

from propan.asyncapi.bindings import (
    AsyncAPIChannelBinding,
    AsyncAPIOperationBinding,
    nats,
)
from propan.asyncapi.channels import AsyncAPIChannel
from propan.asyncapi.message import AsyncAPIMessage
from propan.asyncapi.subscription import AsyncAPISubscription
from propan.brokers._model.schemas import BaseHandler
from propan.types import AnyDict, DecoratedCallable


@dataclass
class Handler(BaseHandler):
    subject: str
    extra_args: AnyDict

    queue: str = ""
    subscription: Optional[Subscription] = None

    def __init__(
        self,
        callback: DecoratedCallable,
        dependant: CallModel,
        subject: str,
        queue: str = "",
        subscription: Optional[Subscription] = None,
        extra_args: Optional[AnyDict] = None,
        _description: str = "",
    ):
        self.callback = callback
        self.dependant = dependant
        self._description = _description
        self.subject = subject
        self.queue = queue
        self.extra_args = extra_args or {}
        self.subscription = subscription

    def get_schema(self) -> Dict[str, AsyncAPIChannel]:
        message_title, body, reply_to = self.get_message_object()

        return {
            self.title: AsyncAPIChannel(
                subscribe=AsyncAPISubscription(
                    description=self.description,
                    bindings=AsyncAPIOperationBinding(
                        nats=nats.AsyncAPINatsOperationBinding(
                            replyTo=reply_to,
                        ),
                    ),
                    message=AsyncAPIMessage(
                        title=message_title,
                        payload=body,
                    ),
                ),
                bindings=AsyncAPIChannelBinding(
                    nats=nats.AsyncAPINatsChannelBinding(
                        subject=self.subject,
                        queue=self.queue or None,
                    )
                ),
            ),
        }



from pathlib import Path

from propan.cli.startproject.async_app.core import create_app_file, create_handlers_file
from propan.cli.startproject.core import (
    create_apps_dir,
    create_config_dir,
    create_env,
    create_project_dir,
)
from propan.cli.startproject.utils import touch_dir, write_file


def create_kafka(dir: Path) -> Path:
    project_dir = _create_project_dir(dir)
    app_dir = _create_app_dir(project_dir / "app")
    _create_config_dir(app_dir / "config")
    _create_apps_dir(app_dir / "apps")
    return project_dir


def _create_project_dir(dirname: Path) -> Path:
    project_dir = create_project_dir(dirname, "propan[async-kafka]")

    write_file(
        project_dir / "docker-compose.yaml",
        'version: "3"',
        "",
        "services:",
        "  kafka:",
        "    image: bitnami/kafka",
        "    ports:",
        "      - 9092:9092",
        "    environment:",
        "      - KAFKA_ENABLE_KRAFT=yes",
        "      - KAFKA_CFG_NODE_ID=1",
        "      - KAFKA_CFG_PROCESS_ROLES=broker,controller",
        "      - KAFKA_CFG_CONTROLLER_LISTENER_NAMES=CONTROLLER",
        "      - KAFKA_CFG_LISTENERS=PLAINTEXT://:9092,CONTROLLER://:9093",
        "      - KAFKA_CFG_LISTENER_SECURITY_PROTOCOL_MAP=CONTROLLER:PLAINTEXT,PLAINTEXT:PLAINTEXT",
        "      - KAFKA_CFG_ADVERTISED_LISTENERS=PLAINTEXT://127.0.0.1:9092",
        "      - KAFKA_BROKER_ID=1",
        "      - KAFKA_CFG_CONTROLLER_QUORUM_VOTERS=1@kafka:9093",
        "      - ALLOW_PLAINTEXT_LISTENER=yes",
        "",
        "  app:",
        "    build: .",
        "    environment:",
        "      APP_BROKER__URL: kafka:9092",
        "    volumes:",
        "      - ./app:/home/user/app:ro",
        "    depends_on:",
        "      - kafka",
    )

    return project_dir


def _create_app_dir(app: Path) -> Path:
    app_dir = touch_dir(app)
    create_app_file(app_dir, "KafkaBroker")
    return app_dir


def _create_config_dir(config: Path) -> Path:
    config_dir = create_config_dir(config)
    create_env(config_dir, "localhost:9092")
    return config_dir


def _create_apps_dir(apps: Path) -> Path:
    apps_dir = create_apps_dir(apps)

    create_handlers_file(
        apps_dir / "handlers.py",
        "KafkaRouter",
    )

    return apps_dir



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from propan.asyncapi.utils import AsyncAPIExternalDocs, AsyncAPITag


class AsyncAPICorrelationId(BaseModel):
    description: Optional[str] = None
    location: str


class AsyncAPIMessage(BaseModel):
    title: Optional[str] = None
    name: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    message_id: Optional[str] = Field(
        default=None,
        alias="messageId",
    )
    correlation_id: Optional[AsyncAPICorrelationId] = Field(
        default=None,
        alias="correlationId",
    )
    content_type: Optional[str] = Field(
        default=None,
        alias="contentType",
    )

    payload: Dict[str, Any]
    # TODO:
    # headers
    # schemaFormat
    # bindings
    # examples
    # traits

    tags: Optional[List[AsyncAPITag]] = None
    external_docs: Optional[AsyncAPIExternalDocs] = Field(
        default=None,
        alias="externalDocs",
    )



from fast_depends.library import CustomField

from propan.types import AnyDict


class NoCast(CustomField):  # type: ignore
    def __init__(self) -> None:
        super().__init__(cast=False)

    def use(self, **kwargs: AnyDict) -> AnyDict:
        return kwargs



import asyncio
import logging
from functools import partial, wraps
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
from uuid import uuid4

import anyio
from aiokafka import AIOKafkaConsumer, AIOKafkaProducer
from aiokafka.structs import ConsumerRecord
from fast_depends.dependencies import Depends
from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from kafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor
from typing_extensions import Literal, TypeAlias, TypeVar

from propan.__about__ import __version__
from propan.brokers._model.broker_usecase import (
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.exceptions import SkipMessage
from propan.brokers.kafka.schemas import Handler
from propan.brokers.push_back_watcher import BaseWatcher
from propan.types import AnyDict, DecodedMessage, SendableMessage
from propan.utils.context import context

T = TypeVar("T")
CorrelationId: TypeAlias = str
KafkaMessage: TypeAlias = PropanMessage[ConsumerRecord]


class KafkaBroker(
    BrokerAsyncUsecase[ConsumerRecord, Callable[[Tuple[str, ...]], AIOKafkaConsumer]]
):
    _publisher: Optional[AIOKafkaProducer]
    __max_topic_len: int
    response_topic: str
    response_callbacks: Dict[CorrelationId, "asyncio.Future[DecodedMessage]"]
    handlers: List[Handler]

    def __init__(
        self,
        bootstrap_servers: Union[str, List[str]] = "localhost",
        *,
        response_topic: str = "",
        log_fmt: Optional[str] = None,
        protocol: str = "kafka",
        api_version: str = "auto",
        **kwargs: AnyDict,
    ) -> None:
        super().__init__(
            bootstrap_servers,
            log_fmt=log_fmt,
            url_=bootstrap_servers,
            protocol=protocol,
            protocol_version=api_version,
            **kwargs,
        )
        self.__max_topic_len = 4
        self._publisher = None
        self.response_topic = response_topic
        self.response_callbacks = {}

    async def _connect(
        self,
        **kwargs: Any,
    ) -> AIOKafkaConsumer:
        kwargs["client_id"] = kwargs.get("client_id", "propan-" + __version__)

        producer = AIOKafkaProducer(**kwargs)
        context.set_global("producer", producer)
        await producer.start()
        self._publisher = producer
        consumer_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k
            in {
                "bootstrap_servers",
                "loop",
                "client_id",
                "request_timeout_ms",
                "retry_backoff_ms",
                "metadata_max_age_ms",
                "security_protocol",
                "api_version",
                "connections_max_idle_ms",
                "sasl_mechanism",
                "sasl_plain_password",
                "sasl_plain_username",
                "sasl_kerberos_service_name",
                "sasl_kerberos_domain_name",
                "sasl_oauth_token_provider",
            }
            and v
        }
        return partial(AIOKafkaConsumer, **consumer_kwargs)

    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        await super().close(exc_type, exc_val, exec_tb)

        for f in self.response_callbacks.values():
            f.cancel()
        self.response_callbacks = {}

        for handler in self.handlers:
            if handler.task is not None:
                handler.task.cancel()
                handler.task = None

            if handler.consumer is not None:
                await handler.consumer.stop()
                handler.consumer = None

        if self._publisher is not None:
            await self._publisher.stop()
            self._publisher = None

    def handle(
        self,
        *topics: str,
        group_id: Optional[str] = None,
        key_deserializer: Optional[Callable[[bytes], Any]] = None,
        value_deserializer: Optional[Callable[[bytes], Any]] = None,
        fetch_max_wait_ms: int = 500,
        fetch_max_bytes: int = 52428800,
        fetch_min_bytes: int = 1,
        max_partition_fetch_bytes: int = 1 * 1024 * 1024,
        auto_offset_reset: Literal[
            "latest",
            "earliest",
            "none",
        ] = "latest",
        enable_auto_commit: bool = True,
        auto_commit_interval_ms: int = 5000,
        check_crcs: bool = True,
        partition_assignment_strategy: Sequence[AbstractPartitionAssignor] = (
            RoundRobinPartitionAssignor,
        ),
        max_poll_interval_ms: int = 300000,
        rebalance_timeout_ms: Optional[int] = None,
        session_timeout_ms: int = 10000,
        heartbeat_interval_ms: int = 3000,
        consumer_timeout_ms: int = 200,
        max_poll_records: Optional[int] = None,
        exclude_internal_topics: bool = True,
        isolation_level: Literal[
            "read_uncommitted",
            "read_committed",
        ] = "read_uncommitted",
        ## broker
        dependencies: Sequence[Depends] = (),
        description: str = "",
        **original_kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[ConsumerRecord, bool], Awaitable[T_HandlerReturn]],
    ]:
        super().handle()

        def wrapper(
            func: HandlerCallable[T_HandlerReturn],
        ) -> Callable[[ConsumerRecord, bool], Awaitable[T_HandlerReturn]]:
            for t in topics:
                self.__max_topic_len = max((self.__max_topic_len, len(t)))

            func, dependant = self._wrap_handler(
                func,
                extra_dependencies=dependencies,
                **original_kwargs,
            )
            handler = Handler(
                callback=func,
                topics=topics,
                _description=description,
                group_id=group_id,
                consumer_kwargs={
                    "key_deserializer": key_deserializer,
                    "value_deserializer": value_deserializer,
                    "fetch_max_wait_ms": fetch_max_wait_ms,
                    "fetch_max_bytes": fetch_max_bytes,
                    "fetch_min_bytes": fetch_min_bytes,
                    "max_partition_fetch_bytes": max_partition_fetch_bytes,
                    "auto_offset_reset": auto_offset_reset,
                    "enable_auto_commit": enable_auto_commit,
                    "auto_commit_interval_ms": auto_commit_interval_ms,
                    "check_crcs": check_crcs,
                    "partition_assignment_strategy": partition_assignment_strategy,
                    "max_poll_interval_ms": max_poll_interval_ms,
                    "rebalance_timeout_ms": rebalance_timeout_ms,
                    "session_timeout_ms": session_timeout_ms,
                    "heartbeat_interval_ms": heartbeat_interval_ms,
                    "consumer_timeout_ms": consumer_timeout_ms,
                    "max_poll_records": max_poll_records,
                    "exclude_internal_topics": exclude_internal_topics,
                    "isolation_level": isolation_level,
                },
                dependant=dependant,
            )
            self.handlers.append(handler)

            return func

        return wrapper

    async def start(self) -> None:
        if self.response_topic:
            self.handle(
                self.response_topic,
                _raw=True,
                enable_auto_commit=False,
            )(self._consume_response)

        context.set_local(
            "log_context",
            self._get_log_context(None, ""),
        )

        await super().start()

        for handler in self.handlers:  # pragma: no branch
            c = self._get_log_context(None, handler.topics)
            self._log(f"`{handler.callback.__name__}` waiting for messages", extra=c)

            consumer = self._connection(
                *handler.topics,
                group_id=handler.group_id,
                **handler.consumer_kwargs,
            )
            await consumer.start()
            handler.consumer = consumer
            handler.task = asyncio.create_task(self._consume(handler))

    @staticmethod
    async def _parse_message(message: ConsumerRecord) -> KafkaMessage:
        headers = {i: j.decode() for i, j in message.headers}
        return PropanMessage(
            body=message.value,
            raw_message=message,
            message_id=f"{message.offset}-{message.timestamp}",
            reply_to=headers.get("reply_to", ""),
            content_type=headers.get("content-type"),
            headers=headers,
        )

    def _process_message(
        self,
        func: Callable[[KafkaMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[KafkaMessage], Awaitable[T]]:
        @wraps(func)
        async def wrapper(message: KafkaMessage) -> T:
            r = await func(message)

            if message.reply_to:
                await self.publish(
                    message=r or "",
                    headers={"correlation_id": message.headers.get("correlation_id")},
                    topic=message.reply_to,
                )

            return r

        return wrapper

    async def publish(
        self,
        message: SendableMessage,
        topic: str,
        key: Optional[bytes] = None,
        partition: Optional[int] = None,
        timestamp_ms: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
        *,
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = None,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]:
        message, content_type = super()._encode_message(message)

        headers_to_send = {
            "content-type": content_type or "",
            **(headers or {}),
        }

        if callback is True:
            reply_to = reply_to or self.response_topic
            if not reply_to:
                raise ValueError(
                    "You should specify `response_topic` at init or use `reply_to` argument"
                )

        if reply_to:
            correlation_id = str(uuid4())
            headers_to_send.update(
                {
                    "reply_to": reply_to,
                    "correlation_id": correlation_id,
                }
            )
        else:
            correlation_id = ""

        response_future: Optional["asyncio.Future[DecodedMessage]"]
        if callback is True:
            response_future = asyncio.Future()
            self.response_callbacks[correlation_id] = response_future
        else:
            response_future = None

        await self._publisher.send(
            topic=topic,
            value=message,
            key=key,
            partition=partition,
            timestamp_ms=timestamp_ms,
            headers=[(i, j.encode()) for i, j in headers_to_send.items()],
        )

        if response_future is not None:
            if raise_timeout:
                scope = anyio.fail_after
            else:
                scope = anyio.move_on_after

            msg: Any = None
            with scope(callback_timeout):
                msg = await response_future

            if msg:
                return await self._decode_message(msg)

    async def publish_batch(
        self,
        *msgs: SendableMessage,
        topic: str,
        partition: Optional[int] = None,
        timestamp_ms: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        batch = self._publisher.create_batch()

        for msg in msgs:
            message, content_type = super()._encode_message(msg)

            headers_to_send = {
                "content-type": content_type or "",
                **(headers or {}),
            }

            batch.append(
                key=None,
                value=message,
                timestamp=timestamp_ms,
                headers=[(i, j.encode()) for i, j in headers_to_send.items()],
            )

        await self._publisher.send_batch(batch, topic, partition=partition)

    @property
    def fmt(self) -> str:
        return self._fmt or (
            "%(asctime)s %(levelname)s - "
            f"%(topic)-{self.__max_topic_len}s | "
            "%(message_id)-10s "
            "- %(message)s"
        )

    def _get_log_context(
        self,
        message: Optional[KafkaMessage],
        topics: Sequence[str] = (),
    ) -> Dict[str, Any]:
        if topics:
            topic = ", ".join(topics)
        elif message is not None:
            topic = message.raw_message.topic
        else:
            topic = ""

        return {
            "topic": topic,
            **super()._get_log_context(message),
        }

    async def _consume(self, handler: Handler) -> NoReturn:
        c = self._get_log_context(None, handler.topics)

        connected = True
        while True:
            try:
                msg = await handler.consumer.getone()

            except Exception as e:
                if connected is True:
                    self._log(e, logging.WARNING, c)
                    connected = False
                await anyio.sleep(5)

            else:
                if connected is False:
                    self._log("Connection established", logging.INFO, c)
                    connected = True

                await handler.callback(msg)

    async def _consume_response(self, message: KafkaMessage):
        correlation_id = message.headers.get("correlation_id")
        if correlation_id is not None:
            callback = self.response_callbacks.pop(correlation_id, None)
            if callback is not None:
                callback.set_result(message)

        raise SkipMessage()



from aiobotocore.config import AioConfig
from botocore import UNSIGNED

from propan import PropanApp, SQSBroker

broker = SQSBroker(
    "http://localhost:9324/",
    region_name="us-west-2",
    config=AioConfig(signature_version=UNSIGNED),
)
app = PropanApp(broker)


@broker.handle("test")
async def hello(msg: str):
    print(msg)


@app.after_startup
async def pub():
    await broker.publish("hi", "test")



import re
import sys
from contextlib import asynccontextmanager
from types import MethodType
from typing import Any, Optional, Union

from propan.types import AnyDict

if sys.version_info < (3, 8):
    from asyncmock import AsyncMock
else:
    from unittest.mock import AsyncMock

from uuid import uuid4

import aiormq
from aio_pika.message import IncomingMessage
from pamqp import commands as spec
from pamqp.header import ContentHeader

from propan.brokers.rabbit import (
    ExchangeType,
    RabbitBroker,
    RabbitExchange,
    RabbitQueue,
)
from propan.brokers.rabbit.rabbit_broker import PikaSendableMessage, TimeoutType
from propan.brokers.rabbit.utils import validate_exchange, validate_queue
from propan.test.utils import call_handler

__all__ = (
    "build_message",
    "TestRabbitBroker",
)


class PatchedMessage(IncomingMessage):
    @asynccontextmanager
    async def process(self):  # type: ignore
        yield

    async def ack(self, multiple: bool = False) -> None:
        pass

    async def nack(self, multiple: bool = False, requeue: bool = True) -> None:
        pass

    async def reject(self, requeue: bool = False) -> None:
        pass


def build_message(
    message: PikaSendableMessage = "",
    queue: Union[RabbitQueue, str] = "",
    exchange: Union[RabbitExchange, str, None] = None,
    *,
    routing_key: str = "",
    **message_kwargs: AnyDict,
) -> PatchedMessage:
    que, exch = validate_queue(queue), validate_exchange(exchange)
    msg = RabbitBroker._validate_message(message, None, **message_kwargs)

    routing = routing_key or (que.name if que else "")

    return PatchedMessage(
        aiormq.abc.DeliveredMessage(
            delivery=spec.Basic.Deliver(
                exchange=exch.name if exch else "",
                routing_key=routing,
            ),
            header=ContentHeader(
                properties=spec.Basic.Properties(
                    content_type=msg.content_type,
                    message_id=str(uuid4()),
                    headers=msg.headers,
                )
            ),
            body=msg.body,
            channel=AsyncMock(),
        )
    )


async def publish(
    self: RabbitBroker,
    message: PikaSendableMessage = "",
    queue: Union[RabbitQueue, str] = "",
    exchange: Union[RabbitExchange, str, None] = None,
    *,
    routing_key: str = "",
    mandatory: bool = True,
    immediate: bool = False,
    timeout: TimeoutType = None,
    callback: bool = False,
    callback_timeout: Optional[float] = 30.0,
    raise_timeout: bool = False,
    **message_kwargs: AnyDict,
) -> Any:
    exch = validate_exchange(exchange)

    incoming = build_message(
        message=message,
        queue=queue,
        exchange=exch,
        routing_key=routing_key,
        **message_kwargs,
    )

    for handler in self.handlers:  # pragma: no branch
        if handler.exchange == exch:
            call = False

            if exch is None or handler.exchange.type == ExchangeType.DIRECT:
                call = handler.queue.name == incoming.routing_key

            elif handler.exchange.type == ExchangeType.FANOUT:
                call = True

            elif handler.exchange.type == ExchangeType.TOPIC:
                call = re.match(
                    handler.queue.name.replace(".", r"\.").replace("*", ".*"),
                    incoming.routing_key,
                )

            elif handler.exchange.type == ExchangeType.HEADERS:
                queue_headers = handler.queue.bind_arguments
                msg_headers = incoming.headers

                if not queue_headers and not msg_headers:
                    call = True

                else:
                    matcher = queue_headers.pop("x-match", "all")

                    full = True
                    none = True
                    for k, v in queue_headers.items():
                        if msg_headers.get(k) != v:
                            full = False
                        else:
                            none = False

                    if not none:
                        call = matcher == "any" or full

            if call:
                r = await call_handler(
                    handler,
                    incoming,
                    callback,
                    callback_timeout,
                    raise_timeout,
                )
                if callback:  # pragma: no branch
                    return r


def TestRabbitBroker(broker: RabbitBroker) -> RabbitBroker:
    broker._channel = AsyncMock()
    broker.connect = AsyncMock()  # type: ignore
    broker.publish = MethodType(publish, broker)  # type: ignore
    return broker



import asyncio
import logging
from contextlib import suppress
from functools import wraps
from secrets import token_hex
from types import TracebackType
from typing import Any, Awaitable, Callable, Dict, List, Optional, Type, TypeVar

import anyio
import nats.errors
import nats.js
from nats.aio.client import Client
from nats.aio.msg import Msg
from nats.js.client import (
    DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
    DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
    JetStreamContext,
)

from propan.brokers._model.broker_usecase import HandlerCallable, T_HandlerReturn
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.exceptions import WRONG_PUBLISH_ARGS
from propan.brokers.nats import consts as api
from propan.brokers.nats.nats_broker import NatsBroker, NatsMessage
from propan.brokers.nats.schemas import Handler
from propan.brokers.push_back_watcher import (
    BaseWatcher,
    NotPushBackWatcher,
    WatcherContext,
)
from propan.types import AnyDict, DecodedMessage, SendableMessage

T = TypeVar("T")


class NatsJSBroker(NatsBroker):
    _raw_connection: Optional[Client]
    _connection: Optional[JetStreamContext]
    _stream_config: api.StreamConfig

    @property
    def js(self) -> JetStreamContext:
        assert self._connection, "You should start connection first"
        return self._connection

    def __init__(
        self,
        *args: Any,
        # JS
        stream: Optional[str] = "propan-stream",
        description: Optional[str] = None,
        retention: Optional[api.RetentionPolicy] = None,
        max_consumers: Optional[int] = None,
        max_msgs: Optional[int] = None,
        max_bytes: Optional[int] = None,
        discard: Optional[api.DiscardPolicy] = api.DiscardPolicy.OLD,
        max_age: Optional[float] = None,  # in seconds
        max_msgs_per_subject: int = -1,
        max_msg_size: Optional[int] = -1,
        storage: Optional[api.StorageType] = None,
        num_replicas: Optional[int] = None,
        no_ack: bool = False,
        template_owner: Optional[str] = None,
        duplicate_window: int = 0,
        placement: Optional[api.Placement] = None,
        mirror: Optional[api.StreamSource] = None,
        sources: Optional[List[api.StreamSource]] = None,
        sealed: bool = False,
        deny_delete: bool = False,
        deny_purge: bool = False,
        allow_rollup_hdrs: bool = False,
        republish: Optional[api.RePublish] = None,
        allow_direct: Optional[bool] = None,
        mirror_direct: Optional[bool] = None,
        **kwargs: AnyDict,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._raw_connection = None
        self._stream_config = api.StreamConfig(
            name=stream,
            description=description,
            retention=retention,
            max_consumers=max_consumers,
            max_msgs=max_msgs,
            max_bytes=max_bytes,
            discard=discard,
            max_age=max_age,
            max_msgs_per_subject=max_msgs_per_subject,
            max_msg_size=max_msg_size,
            storage=storage,
            num_replicas=num_replicas,
            no_ack=no_ack,
            template_owner=template_owner,
            duplicate_window=duplicate_window,
            placement=placement,
            mirror=mirror,
            sources=sources,
            sealed=sealed,
            deny_delete=deny_delete,
            deny_purge=deny_purge,
            allow_rollup_hdrs=allow_rollup_hdrs,
            republish=republish,
            allow_direct=allow_direct,
            mirror_direct=mirror_direct,
        )

    async def _connect(
        self,
        *,
        url: Optional[str] = None,
        **kwargs: Any,
    ) -> JetStreamContext:
        nc = await super()._connect(url=url, **kwargs)
        stream = nc.jetstream()
        self._raw_connection = nc
        return stream

    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        await super(NatsBroker, self).close(exc_type, exc_val, exec_tb)
        for h in self.handlers:
            if h.subscription is not None:
                await h.subscription.unsubscribe()
                h.subscription = None

        if self._raw_connection is not None:
            await self._raw_connection.drain()
            self._raw_connection = None

    async def _start(self):
        await super()._start()
        try:  # pragma: no branch
            await self._connection.add_stream(
                config=self._stream_config,
                subjects=[h.subject for h in self.handlers],
            )
        except nats.js.errors.BadRequestError as e:
            if (
                e.description
                == "stream name already in use with a different configuration"
            ):
                self._log(e, logging.WARNING)
                await self._connection.update_stream(
                    config=self._stream_config,
                    subjects=[h.subject for h in self.handlers],
                )
            else:  # pragma: no cover
                self._log(e, logging.ERROR)

    def _process_message(
        self,
        func: Callable[[NatsMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher] = None,
    ) -> Callable[[NatsMessage], Awaitable[T]]:
        if watcher is None:  # pragma: no branch
            watcher = NotPushBackWatcher()

        @wraps(func)
        async def wrapper(message: NatsMessage) -> T:
            context = WatcherContext(
                watcher,
                message,
                on_success=message_ack,
                on_error=message_nak,
                on_max=message_term,
            )

            async with context:
                await message.raw_message.in_progress()

                r = await func(message)

                if message.reply_to:
                    reply_msg, content_type = self._encode_message(r)

                    await self._raw_connection.publish(
                        subject=message.reply_to,
                        payload=reply_msg,
                        headers={
                            "content-type": content_type or "",
                        },
                    )

            return r

        return wrapper

    async def _parse_message(self, message: Msg) -> NatsMessage:
        headers = message.header or {}
        return PropanMessage(
            body=message.data,
            content_type=headers.get("content-type", ""),
            headers=headers,
            reply_to=headers.get("reply_to", ""),
            raw_message=message,
        )

    def handle(
        self,
        subject: str,
        queue: str = "",
        *,
        durable: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        ordered_consumer: bool = False,
        idle_heartbeat: Optional[float] = None,
        flow_control: bool = False,
        pending_msgs_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        deliver_policy: Optional[api.DeliverPolicy] = None,
        headers_only: Optional[bool] = None,
        # broker kwargs
        description: str = "",
        **original_kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Msg, bool], Awaitable[T_HandlerReturn]],
    ]:
        super(NatsBroker, self).handle()
        self._max_subject_len = max((self._max_subject_len, len(subject)))
        self._max_queue_len = max((self._max_queue_len, len(queue)))

        def wrapper(func: HandlerCallable) -> HandlerCallable:
            func, dependant = self._wrap_handler(
                func,
                queue=queue,
                subject=subject,
                **original_kwargs,
            )
            handler = Handler(
                callback=func,
                subject=subject,
                queue=queue,
                _description=description,
                dependant=dependant,
                extra_args={
                    "stream": self._stream_config.name,
                    "durable": durable,
                    "manual_ack": True,
                    "ordered_consumer": ordered_consumer,
                    "idle_heartbeat": idle_heartbeat,
                    "flow_control": flow_control,
                    "config": config,
                    "pending_msgs_limit": pending_msgs_limit,
                    "pending_bytes_limit": pending_bytes_limit,
                    "deliver_policy": deliver_policy,
                    "headers_only": headers_only,
                },
            )
            self.handlers.append(handler)

            return func

        return wrapper

    async def publish(
        self,
        message: SendableMessage,
        subject: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]:
        if self._connection is None:
            raise ValueError("NatsConnection not started yet")

        payload, content_type = self._encode_message(message)

        client = self._raw_connection

        if callback is True:
            if reply_to:
                raise WRONG_PUBLISH_ARGS

            token = client._nuid.next()
            token.extend(token_hex(2).encode())
            reply_to = token.decode()

        if reply_to:
            future: asyncio.Future[Msg] = asyncio.Future()
            sub = await client.subscribe(reply_to, future=future, max_msgs=1)
            await sub.unsubscribe(limit=1)

        if raise_timeout:
            scope = anyio.fail_after
        else:
            scope = anyio.move_on_after

        with suppress(nats.errors.TimeoutError):  # py37 compatibility
            with scope(callback_timeout):
                await self._connection.publish(
                    subject=subject,
                    payload=payload,
                    headers={
                        **(headers or {}),
                        "reply_to": reply_to,
                        "content-type": content_type or "",
                    },
                    timeout=callback_timeout,
                    stream=self._stream_config.name,
                )

        if reply_to:
            msg: Any = None
            with scope(callback_timeout):
                msg = await future

            if msg:
                if msg.headers:  # pragma: no branch
                    if (
                        msg.headers.get(nats.js.api.Header.STATUS)
                        == nats.aio.client.NO_RESPONDERS_STATUS
                    ):
                        raise nats.errors.NoRespondersError
                return await self._decode_message(await self._parse_message(msg))


async def message_ack(message: NatsMessage) -> None:
    with suppress(nats.errors.MsgAlreadyAckdError):
        await message.raw_message.ack()


async def message_nak(message: NatsMessage) -> None:
    with suppress(nats.errors.MsgAlreadyAckdError):
        await message.raw_message.nak()


async def message_term(message: NatsMessage) -> None:
    with suppress(nats.errors.MsgAlreadyAckdError):
        await message.raw_message.term()



import importlib.util
import json
import sys
from typing import Any, Dict, List, Type

from fast_depends._compat import PYDANTIC_V2, FieldInfo
from pydantic import BaseModel
from typing_extensions import Never

from propan.types import AnyDict


def is_installed(package: str) -> bool:
    return importlib.util.find_spec(package)


if is_installed("fastapi"):
    from fastapi import __version__ as FASTAPI_VERSION

    FASTAPI_V2 = FASTAPI_VERSION.startswith("0.10")

    if FASTAPI_V2:
        from fastapi._compat import _normalize_errors
        from fastapi.exceptions import ResponseValidationError

        def raise_fastapi_validation_error(errors: List[Any], body: AnyDict) -> Never:
            raise ResponseValidationError(_normalize_errors(errors), body=body)

    else:
        from pydantic import ValidationError as ResponseValidationError
        from pydantic import create_model

        ROUTER_VALIDATION_ERROR_MODEL = create_model("PropanRoute")

        def raise_fastapi_validation_error(errors: List[Any], body: AnyDict) -> Never:
            raise ResponseValidationError(errors, ROUTER_VALIDATION_ERROR_MODEL)


if PYDANTIC_V2:
    from pydantic import ConfigDict
    from pydantic_core import to_jsonable_python as model_to_jsonable

    SCHEMA_FIELD = "json_schema_extra"

    def dump_json(data: Any) -> str:
        return json.dumps(model_to_jsonable(data))

    def get_model_fileds(model: BaseModel) -> Dict[str, FieldInfo]:
        return model.model_fields

    def model_to_json(model: BaseModel, **kwargs: AnyDict) -> str:
        return model.model_dump_json(**kwargs)

    def model_to_dict(model: BaseModel, **kwargs: AnyDict) -> AnyDict:
        return model.model_dump(**kwargs)

    def model_parse(model: Type[BaseModel], data: bytes, **kwargs: AnyDict) -> AnyDict:
        return model.model_validate_json(data, **kwargs)

    def model_schema(model: Type[BaseModel], **kwargs: AnyDict) -> AnyDict:
        return model.model_json_schema(**kwargs)

    def update_model_example(model: Type[BaseModel]) -> Type[BaseModel]:
        config = model.model_config
        schema_extra = config.get("json_schema_extra", {})
        if schema_extra.get("example") is None and sys.version_info >= (3, 8):
            from polyfactory.factories.pydantic_factory import ModelFactory

            factory: BaseModel = type(
                f"{model.__name__}_factory", (ModelFactory,), {"__model__": model}
            ).build()
            schema_extra["example"] = model_to_jsonable(factory)
            config["json_schema_extra"] = schema_extra
        return model

    def model_copy(model: BaseModel, **kwargs: AnyDict) -> AnyDict:
        return model.model_copy(**kwargs)

else:
    from pydantic.config import BaseConfig
    from pydantic.config import ConfigDict as CD
    from pydantic.config import get_config
    from pydantic.json import pydantic_encoder

    # type: ignore[no-redef]
    def ConfigDict(**kwargs: Dict[str, Any]) -> Type[BaseConfig]:
        return get_config(CD(**kwargs))  # type: ignore

    SCHEMA_FIELD = "schema_extra"

    def dump_json(data: Any) -> str:
        return json.dumps(data, default=pydantic_encoder)

    def get_model_fileds(model: BaseModel) -> Dict[str, FieldInfo]:
        return model.__fields__

    def model_to_json(model: BaseModel, **kwargs: AnyDict) -> str:
        return model.json(**kwargs)

    def model_to_dict(model: BaseModel, **kwargs: AnyDict) -> AnyDict:
        return model.dict(**kwargs)

    def model_parse(model: Type[BaseModel], data: bytes, **kwargs: AnyDict) -> AnyDict:
        return model.parse_raw(data, **kwargs)

    def model_schema(model: Type[BaseModel], **kwargs: AnyDict) -> AnyDict:
        return model.schema(**kwargs)

    def model_to_jsonable(model: BaseModel, **kwargs: AnyDict) -> AnyDict:
        return json.loads(model.json(**kwargs))

    def update_model_example(model: Type[BaseModel]) -> Type[BaseModel]:
        schema_extra = model.Config.schema_extra.copy()
        if schema_extra.get("example") is None and sys.version_info >= (3, 8):
            from polyfactory.factories.pydantic_factory import ModelFactory

            factory: BaseModel = type(
                f"{model.__name__}_factory", (ModelFactory,), {"__model__": model}
            ).build()
            schema_extra["example"] = model_to_jsonable(factory)
            return type(
                model.__name__,
                (model,),
                {
                    "Config": type(
                        "Config", (model.Config,), {"schema_extra": schema_extra}
                    )
                },
            )
        else:
            return model

    def model_copy(model: BaseModel, **kwargs: AnyDict) -> AnyDict:
        return model.copy(**kwargs)



from propan.brokers.redis import RedisBroker
from propan.fastapi.core.router import PropanRouter
from propan.types import AnyDict


class RedisRouter(PropanRouter[RedisBroker, AnyDict]):
    broker_class = RedisBroker



from nats.js.api import DeliverPolicy

from propan.brokers.nats.nats_broker import NatsBroker, NatsMessage
from propan.brokers.nats.nats_js_broker import NatsJSBroker
from propan.brokers.nats.routing import NatsRouter

__all__ = (
    "NatsBroker",
    "NatsMessage",
    "DeliverPolicy",
    "NatsJSBroker",
    "NatsRouter",
)



from typing import Awaitable, Callable, Sequence, Union

from fast_depends.dependencies import Depends
from nats.aio.msg import Msg

from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.routing import BrokerRouter

class NatsRouter(BrokerRouter[Msg]):
    def handle(  # type: ignore[override]
        self,
        subject: str,
        queue: str = "",
        *,
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Msg, bool], Awaitable[T_HandlerReturn]],
    ]: ...



from typing import Awaitable, Callable

from propan.brokers._model.broker_usecase import HandlerCallable, T_HandlerReturn
from propan.brokers._model.routing import BrokerRouter
from propan.types import AnyDict


class RedisRouter(BrokerRouter[AnyDict]):
    def handle(  # type: ignore[override]
        self,
        channel: str,
        **kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]:
        return super().handle(self.prefix + channel, **kwargs)



"""Simple and fast framework to create message brokers based microservices"""

from unittest.mock import Mock

__version__ = "0.1.5.12"


INSTALL_MESSAGE = (
    "You should specify using broker!\n"
    "Install it using one of the following commands:\n"
    'pip install "propan[async-rabbit]"\n'
    'pip install "propan[async-nats]"\n'
    'pip install "propan[async-redis]"\n'
    'pip install "propan[async-kafka]"\n'
    'pip install "propan[async-sqs]"'
)


def import_error(error_msg: str) -> Mock:
    return Mock(side_effect=ImportError(error_msg))


INSTALL_RABBIT = import_error(
    (
        "\nYou should install RabbitMQ dependencies"
        '\npip install "propan[async-rabbit]"'
    )
)

INSTALL_KAFKA = import_error(
    ("\nYou should install Kafka dependencies" '\npip install "propan[async-kafka]"')
)

INSTALL_NATS = import_error(
    ("\nYou should install NATS dependencies" '\npip install "propan[async-nats]"')
)

INSTALL_SQS = import_error(
    ("\nYou should install SQS dependencies" '\npip install "propan[async-sqs]"')
)

INSTALL_REDIS = import_error(
    ("\nYou should install RabbitMQ dependencies" '\npip install "propan[async-redis]"')
)



from propan import NatsBroker, PropanApp

broker = NatsBroker("nats://localhost:4222")
app = PropanApp(broker)


@broker.handle("test")
async def hello(msg: str):
    print(msg)


@app.after_startup
async def pub():
    await broker.publish("hi", "test")



from abc import abstractmethod
from typing import Any, Callable, Generic, List

from typing_extensions import ParamSpec, TypeVar

from propan.types import AnyDict, SendableMessage

P_RouteCall = ParamSpec("P_RouteCall")
T_RouteReturn = TypeVar("T_RouteReturn", bound=SendableMessage)
MsgType = TypeVar("MsgType")


class BrokerRoute(Generic[MsgType, T_RouteReturn]):
    call: Callable[P_RouteCall, T_RouteReturn]

    def __init__(
        self,
        call: Callable[P_RouteCall, T_RouteReturn],
        *,
        args: Any,
        kwargs: AnyDict,
    ):
        self.call = call
        self.args = args
        self.kwargs = kwargs

    def __call__(self, msg: MsgType, reraise_exc: bool = False) -> T_RouteReturn:
        return self.call(msg, reraise_exc)


class BrokerRouter(Generic[MsgType]):
    prefix: str
    handlers: List[BrokerRoute[MsgType, Any]]

    def __init__(
        self,
        prefix: str = "",
    ):
        self.prefix = prefix
        self.handlers = []

    @abstractmethod
    def handle(
        self,
        *args: Any,
        **kwargs: AnyDict,
    ) -> Callable[
        [Callable[P_RouteCall, T_RouteReturn]],
        Callable[[MsgType, bool], T_RouteReturn],
    ]:
        def router_hanle_wrapper(
            func: Callable[P_RouteCall, T_RouteReturn]
        ) -> BrokerRoute[MsgType, T_RouteReturn]:
            router = BrokerRoute(
                call=func,
                args=args,
                kwargs=kwargs,
            )
            self.handlers.append(router)
            return router

        return router_hanle_wrapper



from propan import KafkaBroker, PropanApp

broker = KafkaBroker("localhost:9092")
app = PropanApp(broker)


@broker.handle("test-topic", auto_offset_reset="earliest")
async def hello(msg: str):
    print(msg)


@app.after_startup
async def pub():
    await broker.publish("hi", "test-topic")



from pathlib import Path

import typer

from propan.cli.startproject.async_app.kafka import create_kafka
from propan.cli.startproject.async_app.nats import create_nats, create_nats_js
from propan.cli.startproject.async_app.rabbit import create_rabbit
from propan.cli.startproject.async_app.redis import create_redis
from propan.cli.startproject.async_app.sqs import create_sqs

async_app = typer.Typer(pretty_exceptions_short=True)


@async_app.command()
def rabbit(appname: str) -> None:
    """Create an asyncronous RabbiMQ Propan project at [APPNAME] directory"""
    project = create_rabbit(Path.cwd() / appname)
    typer.echo(f"Create an asyncronous RabbiMQ Propan project at: {project}")


@async_app.command()
def redis(appname: str) -> None:
    """Create an asyncronous Redis Propan project at [APPNAME] directory"""
    project = create_redis(Path.cwd() / appname)
    typer.echo(f"Create an asyncronous Redis Propan project at: {project}")


@async_app.command()
def nats(appname: str) -> None:
    """Create an asyncronous Nats Propan project at [APPNAME] directory"""
    project = create_nats(Path.cwd() / appname)
    typer.echo(f"Create an asyncronous Nats Propan project at: {project}")


@async_app.command(name="nats-js")
def nats_js(appname: str) -> None:
    """Create an asyncronous NatsJS Propan project at [APPNAME] directory"""
    project = create_nats_js(Path.cwd() / appname)
    typer.echo(f"Create an asyncronous NatsJS Propan project at: {project}")


@async_app.command()
def kafka(appname: str) -> None:
    """Create an asyncronous Kafka Propan project at [APPNAME] directory"""
    project = create_kafka(Path.cwd() / appname)
    typer.echo(f"Create an asyncronous Kafka Propan project at: {project}")


@async_app.command()
def sqs(appname: str) -> None:
    """Create an asyncronous SQS Propan project at [APPNAME] directory"""
    project = create_sqs(Path.cwd() / appname)
    typer.echo(f"Create an asyncronous SQS Propan project at: {project}")



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from propan.asyncapi.utils import AsyncAPITag


class AsyncAPIServer(BaseModel):
    url: str
    protocol: str
    description: Optional[str] = None
    protocol_version: Optional[str] = Field(
        default=None,
        alias="protocolVersion",
    )
    tags: Optional[List[AsyncAPITag]] = None
    security: Optional[Dict[str, List[str]]] = None

    # TODO:
    # variables
    # bindings



from typing import Any, Awaitable, Callable, Optional, Sequence, Union

from aiokafka.structs import ConsumerRecord
from fast_depends.dependencies import Depends
from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from kafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor
from typing_extensions import Literal

from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.routing import BrokerRouter

class KafkaRouter(BrokerRouter[ConsumerRecord]):
    def handle(  # type: ignore[override]
        self,
        *topics: str,
        group_id: Optional[str] = None,
        key_deserializer: Optional[Callable[[bytes], Any]] = None,
        value_deserializer: Optional[Callable[[bytes], Any]] = None,
        fetch_max_wait_ms: int = 500,
        fetch_max_bytes: int = 52428800,
        fetch_min_bytes: int = 1,
        max_partition_fetch_bytes: int = 1 * 1024 * 1024,
        auto_offset_reset: Literal[
            "latest",
            "earliest",
            "none",
        ] = "latest",
        enable_auto_commit: bool = True,
        auto_commit_interval_ms: int = 5000,
        check_crcs: bool = True,
        partition_assignment_strategy: Sequence[AbstractPartitionAssignor] = (
            RoundRobinPartitionAssignor,
        ),
        max_poll_interval_ms: int = 300000,
        rebalance_timeout_ms: Optional[int] = None,
        session_timeout_ms: int = 10000,
        heartbeat_interval_ms: int = 3000,
        consumer_timeout_ms: int = 200,
        max_poll_records: Optional[int] = None,
        exclude_internal_topics: bool = True,
        isolation_level: Literal[
            "read_uncommitted",
            "read_committed",
        ] = "read_uncommitted",
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[ConsumerRecord] = None,
        parse_message: AsyncParser[ConsumerRecord] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[ConsumerRecord, bool], Awaitable[T_HandlerReturn]],
    ]: ...



import asyncio
import logging
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from aiobotocore.client import AioBaseClient
from aiobotocore.config import AioConfig
from fast_depends.dependencies import Depends
from typing_extensions import TypeAlias

from propan.brokers._model import BrokerAsyncUsecase
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.middlewares import BaseMiddleware
from propan.brokers.push_back_watcher import BaseWatcher
from propan.brokers.sqs.schema import Handler, SQSQueue
from propan.log import access_logger
from propan.types import AnyDict, DecodedMessage, SendableMessage

T = TypeVar("T")
QueueUrl: TypeAlias = str
SQSMessage: TypeAlias = PropanMessage[AnyDict]

class SQSBroker(BrokerAsyncUsecase[AnyDict, AioBaseClient]):
    _queues: Dict[str, QueueUrl]
    response_queue: str
    middlewares: Sequence[Type[BaseMiddleware[AnyDict]]]
    response_callbacks: Dict[str, "asyncio.Future[DecodedMessage]"]
    handlers: List[Handler]

    def __init__(
        self,
        url: str = "http://localhost:9324/",
        *,
        response_queue: str = "",
        region_name: Optional[str] = None,
        api_version: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[bool] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        config: Optional[AioConfig] = None,
        # broker
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        dependencies: Sequence[Depends] = (),
        middlewares: Sequence[Type[BaseMiddleware[AnyDict]]] = (),
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        protocol: str = "sqs",
    ) -> None:
        """"""
    async def connect(
        self,
        *,
        url: str = "http://localhost:9324/",
        region_name: Optional[str] = None,
        api_version: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[bool] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        config: Optional[AioConfig] = None,
    ) -> AioBaseClient:
        """"""
    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        """"""
    async def publish(  # type: ignore[override]
        self,
        message: SendableMessage,
        queue: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        delay_seconds: int = 0,
        message_attributes: Optional[Dict[str, Any]] = None,
        message_system_attributes: Optional[Dict[str, Any]] = None,
        deduplication_id: Optional[str] = None,
        group_id: Optional[str] = None,
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = None,
        raise_timeout: bool = False,
    ) -> None:
        """"""
    def handle(  # type: ignore[override]
        self,
        queue: Union[str, SQSQueue],
        *,
        wait_interval: int = 1,
        max_messages_number: int = 10,  # 1...10
        attributes: Sequence[str] = (),
        message_attributes: Sequence[str] = (),
        request_attempt_id: Optional[str] = None,
        visibility_timeout: int = 0,
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]:
        """"""
    async def start(self) -> None:
        """"""
    async def create_queue(self, queue: str) -> QueueUrl:
        """"""
    async def delete_queue(self, queue: str) -> None:
        """"""
    async def get_queue(self, queue: str) -> QueueUrl:
        """"""
    async def delete_message(self) -> None:
        """"""
    async def _consume(self, queue_url: str, handler: Handler) -> NoReturn: ...
    @property
    def fmt(self) -> str: ...
    def _get_log_context(  # type: ignore[override]
        self, message: Optional[SQSMessage], queue: str
    ) -> Dict[str, Any]: ...
    @classmethod
    def _build_message(
        cls,
        message: SendableMessage,
        queue_url: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        delay_seconds: int = 0,  # 0...900
        message_attributes: Optional[Dict[str, Any]] = None,
        message_system_attributes: Optional[Dict[str, Any]] = None,
        # FIFO only
        deduplication_id: Optional[str] = None,
        group_id: Optional[str] = None,
        # broker
        reply_to: str = "",
    ) -> Dict[str, Any]: ...
    async def _parse_message(self, message: Dict[str, Any]) -> SQSMessage: ...
    def _process_message(
        self,
        func: Callable[[SQSMessage], T],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[SQSMessage], T]: ...
    async def _connect(self, *args: Any, **kwargs: Any) -> AioBaseClient: ...



import logging
from asyncio import AbstractEventLoop
from enum import Enum
from ssl import SSLContext
from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Type, Union

from aiokafka.abc import AbstractTokenProvider
from aiokafka.producer.producer import _missing
from aiokafka.structs import ConsumerRecord
from fastapi import params
from fastapi.datastructures import Default
from fastapi.routing import APIRoute
from fastapi.utils import generate_unique_id
from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from kafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor
from kafka.partitioner.default import DefaultPartitioner
from starlette import routing
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp
from typing_extensions import Literal, TypeVar

from propan import KafkaBroker
from propan.__about__ import __version__
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.fastapi.core import PropanRouter
from propan.log import access_logger

Partition = TypeVar("Partition")

class KafkaRouter(PropanRouter[KafkaBroker]):
    def __init__(
        self,
        bootstrap_servers: Union[str, List[str]] = "localhost",
        *,
        # both
        client_id: str = "propan-" + __version__,
        request_timeout_ms: int = 40 * 1000,
        retry_backoff_ms: int = 100,
        metadata_max_age_ms: int = 5 * 60 * 1000,
        security_protocol: Literal[
            "SSL",
            "PLAINTEXT",
        ] = "PLAINTEXT",
        api_version: str = "auto",
        connections_max_idle_ms: int = 540000,
        sasl_mechanism: Literal[
            "PLAIN",
            "GSSAPI",
            "SCRAM-SHA-256",
            "SCRAM-SHA-512",
            "OAUTHBEARER",
        ] = "PLAIN",
        sasl_plain_password: Optional[str] = None,
        sasl_plain_username: Optional[str] = None,
        sasl_kerberos_service_name: str = "kafka",
        sasl_kerberos_domain_name: Optional[str] = None,
        sasl_oauth_token_provider: Optional[AbstractTokenProvider] = None,
        # publisher
        acks: Union[Literal[0, 1, -1, "all"], object] = _missing,
        key_serializer: Optional[Callable[[Any], bytes]] = None,
        value_serializer: Optional[Callable[[Any], bytes]] = None,
        compression_type: Optional[Literal["gzip", "snappy", "lz4", "zstd"]] = None,
        max_batch_size: int = 16384,
        partitioner: Callable[
            [bytes, List[Partition], List[Partition]],
            Partition,
        ] = DefaultPartitioner(),
        max_request_size: int = 1048576,
        linger_ms: int = 0,
        send_backoff_ms: int = 100,
        ssl_context: Optional[SSLContext] = None,
        enable_idempotence: bool = False,
        transactional_id: Optional[str] = None,
        transaction_timeout_ms: int = 60000,
        # FastAPI kwargs
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[routing.BaseRoute]] = None,
        routes: Optional[List[routing.BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(
            generate_unique_id
        ),
        loop: Optional[AbstractEventLoop] = None,
        # Broker kwargs
        schema_url: str = "/asyncapi",
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        decode_message: AsyncDecoder[ConsumerRecord] = None,
        parse_message: AsyncParser[ConsumerRecord] = None,
        protocol: str = "kafka",
    ) -> None:
        pass
    def add_api_mq_route(  # type: ignore[override]
        self,
        *topics: str,
        endpoint: HandlerCallable[T_HandlerReturn],
        group_id: Optional[str] = None,
        key_deserializer: Optional[Callable[[bytes], Any]] = None,
        value_deserializer: Optional[Callable[[bytes], Any]] = None,
        fetch_max_wait_ms: int = 500,
        fetch_max_bytes: int = 52428800,
        fetch_min_bytes: int = 1,
        max_partition_fetch_bytes: int = 1 * 1024 * 1024,
        auto_offset_reset: Literal[
            "latest",
            "earliest",
            "none",
        ] = "latest",
        enable_auto_commit: bool = True,
        auto_commit_interval_ms: int = 5000,
        check_crcs: bool = True,
        partition_assignment_strategy: Sequence[AbstractPartitionAssignor] = (
            RoundRobinPartitionAssignor,
        ),
        max_poll_interval_ms: int = 300000,
        rebalance_timeout_ms: Optional[int] = None,
        session_timeout_ms: int = 10000,
        heartbeat_interval_ms: int = 3000,
        consumer_timeout_ms: int = 200,
        max_poll_records: Optional[int] = None,
        exclude_internal_topics: bool = True,
        isolation_level: Literal[
            "read_uncommitted",
            "read_committed",
        ] = "read_uncommitted",
        # broker kwargs
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[ConsumerRecord] = None,
        parse_message: AsyncParser[ConsumerRecord] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[[ConsumerRecord, bool], Awaitable[T_HandlerReturn]]:
        pass
    def event(  # type: ignore[override]
        self,
        *topics: str,
        group_id: Optional[str] = None,
        key_deserializer: Optional[Callable[[bytes], Any]] = None,
        value_deserializer: Optional[Callable[[bytes], Any]] = None,
        fetch_max_wait_ms: int = 500,
        fetch_max_bytes: int = 52428800,
        fetch_min_bytes: int = 1,
        max_partition_fetch_bytes: int = 1 * 1024 * 1024,
        auto_offset_reset: Literal[
            "latest",
            "earliest",
            "none",
        ] = "latest",
        enable_auto_commit: bool = True,
        auto_commit_interval_ms: int = 5000,
        check_crcs: bool = True,
        partition_assignment_strategy: Sequence[AbstractPartitionAssignor] = (
            RoundRobinPartitionAssignor,
        ),
        max_poll_interval_ms: int = 300000,
        rebalance_timeout_ms: Optional[int] = None,
        session_timeout_ms: int = 10000,
        heartbeat_interval_ms: int = 3000,
        consumer_timeout_ms: int = 200,
        max_poll_records: Optional[int] = None,
        exclude_internal_topics: bool = True,
        isolation_level: Literal[
            "read_uncommitted",
            "read_committed",
        ] = "read_uncommitted",
        # broker kwargs
        dependencies: Optional[Sequence[params.Depends]] = None,
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[ConsumerRecord] = None,
        parse_message: AsyncParser[ConsumerRecord] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[ConsumerRecord, bool], Awaitable[T_HandlerReturn]],
    ]: ...



from message_pb2 import Person

from propan import PropanApp, RabbitBroker
from propan.annotations import Logger, NoCast
from propan.brokers.rabbit import RabbitMessage

broker = RabbitBroker()
app = PropanApp(broker)


async def decode_message(msg: RabbitMessage, *args) -> Person:
    decoded = Person()
    decoded.ParseFromString(msg.body)
    return decoded


@broker.handle("test", decode_message=decode_message)
async def consume(body: NoCast[Person], logger: Logger):
    logger.info(body)


@app.after_startup
async def publish():
    body = Person(name="john", age=25).SerializeToString()
    await broker.publish(body, "test")



import inspect
from functools import wraps
from typing import Awaitable, Callable, List, TypeVar, Union, cast

from fast_depends.utils import run_async as call_or_await
from typing_extensions import ParamSpec

__all__ = (
    "call_or_await",
    "to_async",
)

T = TypeVar("T")
P = ParamSpec("P")


def to_async(
    func: Union[
        Callable[P, T],
        Callable[P, Awaitable[T]],
    ]
) -> Callable[P, Awaitable[T]]:
    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        r = await call_or_await(func, *args, **kwargs)
        return cast(T, r)

    return wrapper


def get_function_positional_arguments(func: Callable[P, T]) -> List[str]:
    signature = inspect.signature(func)

    arg_kinds = (
        inspect.Parameter.POSITIONAL_ONLY,
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
    )

    args = [
        param.name for param in signature.parameters.values() if param.kind in arg_kinds
    ]

    return args



import json
import sys
from hashlib import md5
from types import MethodType
from typing import Any, Dict, Optional
from uuid import uuid4

if sys.version_info < (3, 8):
    from asyncmock import AsyncMock
else:
    from unittest.mock import AsyncMock

from propan import SQSBroker
from propan.brokers.sqs.schema import SQSMessage
from propan.test.utils import call_handler
from propan.types import SendableMessage

__all__ = (
    "build_message",
    "TestSQSBroker",
)


def build_message(
    message: SendableMessage,
    queue: str = "",
    headers: Optional[Dict[str, str]] = None,
    delay_seconds: int = 0,  # 0...900
    message_attributes: Optional[Dict[str, Any]] = None,
    message_system_attributes: Optional[Dict[str, Any]] = None,
    # FIFO only
    deduplication_id: Optional[str] = None,
    group_id: Optional[str] = None,
    reply_to: str = "",
) -> Dict[str, Any]:
    params = SQSMessage(
        message=message,
        delay_seconds=delay_seconds,
        headers=headers or {},
        message_attributes=message_attributes or {},
        message_system_attributes=message_system_attributes or {},
        deduplication_id=deduplication_id,
        group_id=group_id,
    ).to_params(reply_to=reply_to)

    body = params.get("MessageBody", "0")
    attributes = params.get("MessageAttributes", {})

    return {
        "Body": body,
        "MD5OfBody": md5(body.encode()).hexdigest(),
        "MD5OfMessageAttributes": md5(json.dumps(attributes).encode()).hexdigest(),
        "MessageAttributes": attributes,
        "MessageId": str(uuid4()),
        "ReceiptHandle": str(uuid4()),
    }


async def publish(
    self: SQSBroker,
    message: SendableMessage,
    queue: str,
    *,
    headers: Optional[Dict[str, str]] = None,
    delay_seconds: int = 0,  # 0...900
    message_attributes: Optional[Dict[str, Any]] = None,
    message_system_attributes: Optional[Dict[str, Any]] = None,
    # FIFO only
    deduplication_id: Optional[str] = None,
    group_id: Optional[str] = None,
    # broker
    reply_to: str = "",
    callback: bool = False,
    callback_timeout: Optional[float] = None,
    raise_timeout: bool = False,
) -> Any:
    incoming = build_message(
        message=message,
        headers=headers,
        delay_seconds=delay_seconds,
        message_attributes=message_attributes,
        message_system_attributes=message_system_attributes,
        deduplication_id=deduplication_id,
        group_id=group_id,
        reply_to=reply_to,
    )

    for handler in self.handlers:  # pragma: no branch
        if queue == handler.queue.name:  # pragma: no branch
            r = await call_handler(
                handler, incoming, callback, callback_timeout, raise_timeout
            )
            if callback:  # pragma: no branch
                return r


async def delete_message(self: SQSBroker) -> None:
    pass


def TestSQSBroker(broker: SQSBroker) -> SQSBroker:
    broker.connect = AsyncMock()  # type: ignore
    broker.start = AsyncMock()  # type: ignore
    broker.delete_message = MethodType(delete_message, broker)  # type: ignore
    broker.publish = MethodType(publish, broker)  # type: ignore
    return broker



from typing import Optional, Union, overload

from propan.brokers.rabbit.schemas import RabbitExchange, RabbitQueue


@overload
def validate_exchange(
    exchange: None = None,
) -> None:
    ...


@overload
def validate_exchange(
    exchange: Union[str, RabbitExchange],
) -> RabbitExchange:
    ...


def validate_exchange(
    exchange: Union[str, RabbitExchange, None] = None,
) -> Optional[RabbitExchange]:
    if exchange is not None:  # pragma: no branch
        if isinstance(exchange, str):
            exchange = RabbitExchange(name=exchange)
        elif not isinstance(exchange, RabbitExchange):
            raise ValueError(
                f"Exchange '{exchange}' should be 'str' | 'RabbitExchange' instance"
            )
    return exchange


@overload
def validate_queue(
    queue: None = None,
) -> None:
    ...


@overload
def validate_queue(
    queue: Union[str, RabbitQueue],
) -> RabbitQueue:
    ...


def validate_queue(
    queue: Union[str, RabbitQueue, None] = None
) -> Optional[RabbitQueue]:
    if queue is not None:  # pragma: no branch
        if isinstance(queue, str):
            queue = RabbitQueue(name=queue)
        elif not isinstance(queue, RabbitQueue):
            raise ValueError(
                f"Queue '{queue}' should be 'str' | 'RabbitQueue' instance"
            )
    return queue



import logging
import sys
from collections import defaultdict
from types import TracebackType
from typing import Callable, DefaultDict, Mapping, Optional, Tuple, Type, Union

import click
from typing_extensions import Literal

from propan.utils.context.main import context

original_makeRecord = logging.Logger.makeRecord


class ColourizedFormatter(logging.Formatter):
    level_name_colors: DefaultDict[str, Callable[[str], str]] = defaultdict(
        lambda: str,
        **{
            str(logging.DEBUG): lambda level_name: click.style(
                str(level_name), fg="cyan"
            ),
            str(logging.INFO): lambda level_name: click.style(
                str(level_name), fg="green"
            ),
            str(logging.WARNING): lambda level_name: click.style(
                str(level_name), fg="yellow"
            ),
            str(logging.ERROR): lambda level_name: click.style(
                str(level_name), fg="red"
            ),
            str(logging.CRITICAL): lambda level_name: click.style(
                str(level_name), fg="bright_red"
            ),
        },
    )

    def __init__(
        self,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        style: Literal["%", "{", "$"] = "%",
        use_colors: Optional[bool] = None,
    ):
        """
        Initialize the formatter with specified format strings.

        Initialize the formatter either with the specified format string, or a
        default as described above. Allow for specialized date formatting with
        the optional datefmt argument. If datefmt is omitted, you get an
        ISO8601-like (or RFC 3339-like) format.

        Use a style parameter of '%', '{' or '$' to specify that you want to
        use one of %-formatting, :meth:`str.format` (``{}``) formatting or
        :class:`string.Template` formatting in your format string.
        """
        if use_colors in (True, False):
            self.use_colors = use_colors
        else:
            self.use_colors = sys.stdout.isatty()
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)

    def color_level_name(self, level_name: str, level_no: int) -> str:
        return self.level_name_colors[str(level_no)](level_name)

    def formatMessage(self, record: logging.LogRecord) -> str:
        levelname = expand_log_field(record.levelname, 8)
        if self.use_colors is True:  # pragma: no cover
            levelname = self.color_level_name(levelname, record.levelno)
        record.__dict__["levelname"] = levelname
        return super().formatMessage(record)


def make_record_with_extra(
    self: logging.Logger,
    name: str,
    level: int,
    fn: str,
    lno: int,
    msg: str,
    args: Tuple[str],
    exc_info: Optional[
        Union[
            Tuple[Type[BaseException], BaseException, Optional[TracebackType]],
            Tuple[None, None, None],
        ]
    ],
    func: Optional[str] = None,
    extra: Optional[Mapping[str, object]] = None,
    sinfo: Optional[str] = None,
) -> logging.LogRecord:
    if extra is None:
        extra = context.get_local("log_context")

    record = original_makeRecord(
        self, name, level, fn, lno, msg, args, exc_info, func, extra, sinfo
    )

    return record


def expand_log_field(field: str, symbols: int) -> str:
    return field + (" " * (symbols - len(field)))


logging.Logger.makeRecord = make_record_with_extra  # type: ignore



from typing import Awaitable, Callable, Sequence

from fast_depends.dependencies import Depends

from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.routing import BrokerRouter
from propan.types import AnyDict

class RedisRouter(BrokerRouter[AnyDict]):
    def handle(  # type: ignore[override]
        self,
        channel: str,
        *,
        pattern: bool = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]: ...



from propan import PropanApp, RedisBroker
from propan.annotations import Logger

broker = RedisBroker()
app = PropanApp(broker)


@broker.handle("test")
async def handler1(logger: Logger):
    logger.info("handler1")


@broker.handle("test")
async def handler2(logger: Logger):
    logger.info("handler2")


@broker.handle("test2")
async def handler3(logger: Logger):
    logger.info("handler3")


@app.after_startup
async def publish_smth():
    await broker.publish("", "test")
    await broker.publish("", "test2")



import asyncio
import logging
from functools import wraps
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    Sequence,
    Type,
    TypeVar,
)
from uuid import uuid4

import anyio
from anyio.streams.memory import MemoryObjectSendStream
from fast_depends.dependencies import Depends
from redis.asyncio.client import PubSub, Redis
from redis.asyncio.connection import ConnectionPool, parse_url
from typing_extensions import TypeAlias

from propan._compat import model_parse, model_to_json
from propan.brokers._model.broker_usecase import (
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage, RawDecoced
from propan.brokers.push_back_watcher import BaseWatcher
from propan.brokers.redis.schemas import Handler
from propan.brokers.redis.schemas import RedisMessage as RM
from propan.types import AnyDict, DecodedMessage, SendableMessage
from propan.utils import context

T = TypeVar("T")
RedisMessage: TypeAlias = PropanMessage[AnyDict]


class RedisBroker(BrokerAsyncUsecase):
    handlers: List[Handler]
    _connection: Redis
    __max_channel_len: int
    _polling_interval: float

    def __init__(
        self,
        url: str = "redis://localhost:6379",
        *,
        polling_interval: float = 1.0,
        log_fmt: Optional[str] = None,
        protocol: str = "redis",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            url,
            log_fmt=log_fmt,
            url_=url,
            protocol=protocol,
            **kwargs,
        )
        self.__max_channel_len = 0
        self._polling_interval = polling_interval

    async def _connect(
        self,
        url: str,
        **kwargs: Any,
    ) -> Redis:
        url_options = parse_url(url)
        url_options.update(kwargs)
        pool = ConnectionPool(**url_options)
        return Redis(connection_pool=pool)

    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        await super().close(exc_type, exc_val, exec_tb)
        for h in self.handlers:
            if h.task is not None:  # pragma: no branch
                h.task.cancel()
                h.task = None

            if h.subscription is not None:  # pragma: no branch
                await h.subscription.unsubscribe()
                await h.subscription.reset()
                h.subscription = None

        if self._connection is not None:  # pragma: no branch
            await self._connection.close()
            self._connection = None

    def _process_message(
        self,
        func: Callable[[RedisMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[RedisMessage], Awaitable[T]]:
        @wraps(func)
        async def wrapper(message: RedisMessage) -> T:
            r = await func(message)

            msg = message.raw_message
            if isinstance(msg, RM) and message.reply_to:
                await self.publish(r or "", message.reply_to)

            return r

        return wrapper

    def handle(
        self,
        channel: str = "",
        *,
        pattern: bool = False,
        dependencies: Sequence[Depends] = (),
        description: str = "",
        **original_kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]:
        super().handle()

        self.__max_channel_len = max(self.__max_channel_len, len(channel))

        def wrapper(
            func: HandlerCallable[T_HandlerReturn],
        ) -> Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]]:
            func, dependant = self._wrap_handler(
                func,
                channel=channel,
                extra_dependencies=dependencies,
                **original_kwargs,
            )
            handler = Handler(
                callback=func,
                channel=channel,
                pattern=pattern,
                _description=description,
                dependant=dependant,
            )
            self.handlers.append(handler)

            return func

        return wrapper

    async def start(self) -> None:
        context.set_local(
            "log_context",
            self._get_log_context(None, ""),
        )

        await super().start()

        for handler in self.handlers:  # pragma: no branch
            c = self._get_log_context(None, handler.channel)
            self._log(f"`{handler.callback.__name__}` waiting for messages", extra=c)

            psub = self._connection.pubsub()
            if handler.pattern is True:
                await psub.psubscribe(handler.channel)
            else:
                await psub.subscribe(handler.channel)

            handler.subscription = psub
            handler.task = asyncio.create_task(self._consume(handler, psub))

    async def publish(
        self,
        message: SendableMessage = "",
        channel: str = "",
        *,
        reply_to: str = "",
        headers: Optional[Dict[str, Any]] = None,
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]:
        if self._connection is None:
            raise ValueError("Redis connection not established yet")

        msg, content_type = self._encode_message(message)

        if callback is True:
            if reply_to:
                raise ValueError(
                    "You should use `reply_to` to send response to long-living queue "
                    "and `callback` to get response in sync mode."
                )

            callback_channel = str(uuid4())
            psub = self._connection.pubsub()
            (
                send_response_stream,
                receive_response_stream,
            ) = anyio.create_memory_object_stream(max_buffer_size=1)
            await psub.subscribe(callback_channel)
            task = asyncio.create_task(_consume_one(send_response_stream, psub))
        else:
            callback_channel = reply_to
            psub = None
            send_response_stream = receive_response_stream = None
            task = None

        await self._connection.publish(
            channel,
            model_to_json(
                RM(
                    data=msg,
                    headers={
                        "content-type": content_type or "",
                        **(headers or {}),
                    },
                    reply_to=callback_channel,
                )
            ),
        )

        if psub and receive_response_stream and task:
            try:
                with anyio.fail_after(callback_timeout):
                    response = await receive_response_stream.receive()
            except TimeoutError as e:
                if raise_timeout is True:
                    raise e
                return None
            else:
                return await self._decode_message(await self._parse_message(response))
            finally:
                await psub.unsubscribe(callback_channel)
                await psub.reset()
                task.cancel()

    @staticmethod
    async def _parse_message(message: Any) -> RedisMessage:
        data = message.get("data", b"")

        try:
            obj = model_parse(RM, data)
        except Exception:
            msg = RedisMessage(
                body=data,
                raw_message=message,
            )
        else:
            msg = RedisMessage(
                body=obj.data,
                content_type=obj.headers.get("content-type", ""),
                reply_to=obj.reply_to,
                headers=obj.headers,
                raw_message=obj,
            )

        return msg

    async def _decode_message(self, message: RedisMessage) -> DecodedMessage:
        if message.headers.get("content-type") is not None:
            return await super()._decode_message(message)
        else:
            return RawDecoced(message=message.body).message

    def _get_log_context(
        self, message: Optional[RedisMessage], channel: str
    ) -> Dict[str, Any]:
        context = {
            "channel": channel,
            **super()._get_log_context(message),
        }
        return context

    @property
    def fmt(self) -> str:
        return self._fmt or (
            "%(asctime)s %(levelname)s - "
            f"%(channel)-{self.__max_channel_len}s | "
            "%(message_id)-10s "
            "- %(message)s"
        )

    async def _consume(self, handler: Handler, psub: PubSub) -> NoReturn:
        c = self._get_log_context(None, handler.channel)

        connected = True
        while True:
            try:
                m = await psub.get_message(
                    ignore_subscribe_messages=True,
                    timeout=self._polling_interval,
                )
            except Exception:
                if connected is True:
                    self._log("Connection broken", logging.WARNING, c)
                    connected = False
                await anyio.sleep(5)
            else:
                if connected is False:
                    self._log("Connection established", logging.INFO, c)
                    connected = True

                if m:  # pragma: no branch
                    await handler.callback(m)
            finally:
                await anyio.sleep(0.01)


async def _consume_one(queue: MemoryObjectSendStream[any], psub: PubSub) -> NoReturn:
    async for m in psub.listen():
        t = m.get("type")
        if t and "message" in t:  # pragma: no branch
            await queue.send(m)
            break



import logging
import ssl
from types import TracebackType
from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Type, Union

from fast_depends.dependencies import Depends
from nats.aio.client import (
    DEFAULT_CONNECT_TIMEOUT,
    DEFAULT_DRAIN_TIMEOUT,
    DEFAULT_INBOX_PREFIX,
    DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
    DEFAULT_MAX_OUTSTANDING_PINGS,
    DEFAULT_MAX_RECONNECT_ATTEMPTS,
    DEFAULT_PENDING_SIZE,
    DEFAULT_PING_INTERVAL,
    DEFAULT_RECONNECT_TIME_WAIT,
    Callback,
    Client,
    Credentials,
    ErrorCallback,
    JWTCallback,
    SignatureCallback,
)
from nats.aio.msg import Msg
from nats.aio.subscription import (
    DEFAULT_SUB_PENDING_BYTES_LIMIT,
    DEFAULT_SUB_PENDING_MSGS_LIMIT,
)
from typing_extensions import TypeAlias

from propan.brokers._model import BrokerAsyncUsecase
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.middlewares import BaseMiddleware
from propan.brokers.nats.schemas import Handler
from propan.brokers.push_back_watcher import BaseWatcher
from propan.log import access_logger
from propan.types import DecodedMessage, SendableMessage

NatsMessage: TypeAlias = PropanMessage[Msg]

class NatsBroker(BrokerAsyncUsecase[Msg, Client]):
    logger: logging.Logger
    middlewares: Sequence[Type[BaseMiddleware[Msg]]]
    handlers: List[Handler]

    def __init__(
        self,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        *,
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
        # broker kwargs
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        dependencies: Sequence[Depends] = (),
        middlewares: Sequence[Type[BaseMiddleware[Msg]]] = (),
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        protocol: str = "nats",
    ) -> None: ...
    async def connect(
        self,
        *,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
    ) -> Client: ...
    async def publish(  # type: ignore[override]
        self,
        message: SendableMessage,
        subject: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]: ...
    def handle(  # type: ignore[override]
        self,
        subject: str,
        queue: str = "",
        *,
        pending_msgs_limit: int = DEFAULT_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = DEFAULT_SUB_PENDING_BYTES_LIMIT,
        # broker kwargs
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Msg, bool], Awaitable[T_HandlerReturn]],
    ]: ...
    def _get_log_context(  # type: ignore[override]
        self,
        message: Optional[NatsMessage],
        subject: str,
        queue: str = "",
    ) -> Dict[str, Any]: ...
    async def _connect(
        self,
        *,
        url: Optional[str] = None,
        error_cb: Optional[ErrorCallback] = None,
        reconnected_cb: Optional[Callback] = None,
        **kwargs: Any,
    ) -> Client: ...
    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None: ...
    async def _parse_message(self, message: Msg) -> NatsMessage: ...
    def _process_message(
        self,
        func: Callable[[NatsMessage], Awaitable[T_HandlerReturn]],
        watcher: Optional[BaseWatcher] = None,
    ) -> Callable[[NatsMessage], Awaitable[T_HandlerReturn]]: ...



import sys
from types import MethodType
from typing import Any, Dict, Optional, Union
from uuid import uuid4

from nats.aio.msg import Msg

if sys.version_info < (3, 8):
    from asyncmock import AsyncMock
else:
    from unittest.mock import AsyncMock

from propan import NatsBroker
from propan.test.utils import call_handler
from propan.types import SendableMessage

__all__ = (
    "build_message",
    "TestNatsBroker",
)


class PatchedMessage(Msg):
    async def ack(self) -> None:
        pass

    async def ack_sync(self, timeout: float = 1) -> None:
        pass

    async def nak(self, delay: Union[int, float, None] = None) -> None:
        pass

    async def term(self) -> None:
        pass

    async def in_progress(self) -> None:
        pass


def build_message(
    message: SendableMessage,
    subject: str,
    *,
    reply_to: str = "",
    headers: Optional[Dict[str, Any]] = None,
) -> PatchedMessage:
    msg, content_type = NatsBroker._encode_message(message)
    return PatchedMessage(
        _client=None,  # type: ignore
        subject=subject,
        reply=reply_to or str(uuid4()),
        data=msg,
        headers={
            **(headers or {}),
            "content-type": content_type or "",
        },
    )


async def publish(
    self: NatsBroker,
    message: SendableMessage,
    subject: str,
    *,
    reply_to: str = "",
    headers: Optional[Dict[str, Any]] = None,
    callback: bool = False,
    callback_timeout: Optional[float] = 30.0,
    raise_timeout: bool = False,
) -> Any:
    incoming = build_message(
        message=message,
        subject=subject,
        reply_to=reply_to,
        headers=headers,
    )

    for handler in self.handlers:  # pragma: no branch
        if subject == handler.subject:  # pragma: no branch
            r = await call_handler(
                handler, incoming, callback, callback_timeout, raise_timeout
            )
            if callback:  # pragma: no branch
                return r


def TestNatsBroker(broker: NatsBroker) -> NatsBroker:
    broker.connect = AsyncMock()  # type: ignore
    broker.start = AsyncMock()  # type: ignore
    broker._raw_connection = AsyncMock()  # type: ignore
    broker.publish = MethodType(publish, broker)  # type: ignore
    return broker



from typing import Any, Optional

import anyio

from propan.brokers._model.schemas import BaseHandler


async def call_handler(
    handler: BaseHandler,
    message: Any,
    callback: bool = False,
    callback_timeout: Optional[float] = 30.0,
    raise_timeout: bool = False,
) -> Any:
    if raise_timeout:
        scope = anyio.fail_after
    else:
        scope = anyio.move_on_after

    result: Any = None
    with scope(callback_timeout):
        result = await handler.callback(message)

    if callback is True:  # pragma: no branch
        return result



from aiokafka.structs import ConsumerRecord

from propan.brokers.kafka import KafkaBroker
from propan.fastapi.core.router import PropanRouter


class KafkaRouter(PropanRouter[KafkaBroker, ConsumerRecord]):
    broker_class = KafkaBroker
